# -*- coding: utf-8 -*-
"""
Exclusion layers handler
"""
import json
import logging

import numpy as np
from rex.multi_file_resource import MultiFileResource
from rex.resource import Resource
from rex.utilities.parse_keys import parse_keys

from reV.utilities.exceptions import HandlerKeyError, MultiFileExclusionError

logger = logging.getLogger(__name__)
LATITUDE, LONGITUDE = "latitude", "longitude"


class ExclusionLayers:
    """
    Handler of .h5 file and techmap for Exclusion Layers
    """

    def __init__(self, h5_file, hsds=False):
        """
        Parameters
        ----------
        h5_file : str | list | tuple
            .h5 file containing exclusion layers and techmap,
            or a list of h5 files
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS
        """

        self.h5_file = h5_file

        if isinstance(h5_file, str):
            self._h5 = Resource(h5_file, hsds=hsds)
        elif isinstance(h5_file, (list, tuple)):
            self._h5 = MultiFileResource(h5_file, check_files=False)
            self._preflight_multi_file()
        else:
            msg = ('Expected str, list, or tuple for h5_file input but '
                   'received {}'.format(type(h5_file)))
            logger.error(msg)
            raise TypeError(msg)

        self._iarr = None

    def __repr__(self):
        msg = "{} for {}".format(self.__class__.__name__, self.h5_file)

        return msg

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

        if type is not None:
            raise

    def __len__(self):
        return len(self.layers)

    def __getitem__(self, keys):
        ds, ds_slice = parse_keys(keys)

        if ds.lower().startswith('lat'):
            out = self._get_latitude(*ds_slice)
        elif ds.lower().startswith('lon'):
            out = self._get_longitude(*ds_slice)
        else:
            out = self._get_layer(ds, *ds_slice)

        return out

    def __contains__(self, layer):
        return layer in self.layers

    def _preflight_multi_file(self):
        """Run simple multi-file exclusion checks."""
        lat_shape = self.h5.shapes[LATITUDE]
        lon_shape = self.h5.shapes[LONGITUDE]
        for layer in self.layers:
            lshape = self.h5.shapes[layer]
            lshape = lshape[1:] if len(lshape) > 2 else lshape
            if lshape != lon_shape or lshape != lat_shape:
                msg = ('Shape of layer "{}" is {} which does not match '
                       'latitude and longitude shapes of {} and {}. '
                       'Check your exclusion file inputs: {}'
                       .format(layer, self.h5.shapes[layer],
                               lat_shape, lon_shape, self.h5._h5_files))
                logger.error(msg)
                raise MultiFileExclusionError(msg)

        check_attrs = ('height', 'width', 'crs', 'transform')
        base_profile = {}
        for fp in self.h5_file:
            with ExclusionLayers(fp) as f:
                if not base_profile:
                    base_profile = f.profile
                else:
                    for attr in check_attrs:
                        if attr not in base_profile or attr not in f.profile:
                            msg = ('Multi-file exclusion inputs from {} '
                                   'dont have profiles with height, width, '
                                   'crs, and transform: {} and {}'
                                   .format(self.h5_file, base_profile,
                                           f.profile))
                            logger.error(msg)
                            raise MultiFileExclusionError(msg)

                        base_attr = base_profile[attr]
                        file_attr = f.profile[attr]
                        attrs_are_str = (isinstance(base_attr, str)
                                         and isinstance(file_attr, str))
                        if attr == 'crs' and attrs_are_str:
                            attrs_match = (set(base_attr.split(' '))
                                           == set(file_attr.split(' ')))
                        else:
                            attrs_match = base_profile[attr] == f.profile[attr]

                        if not attrs_match:
                            msg = ('Multi-file exclusion inputs from {} '
                                   'dont have matching "{}": {} and {}'
                                   .format(self.h5_file, attr,
                                           base_profile[attr],
                                           f.profile[attr]))
                            logger.error(msg)
                            raise MultiFileExclusionError(msg)

    def close(self):
        """
        Close h5 instance
        """
        self._h5.close()

    @property
    def h5(self):
        """
        Open h5py File instance.

        Returns
        -------
        h5 : rex.MultiFileResource | rex.Resource
        """
        return self._h5

    @property
    def iarr(self):
        """Get an array of 1D index values for the flattened h5 excl extent.

        Returns
        -------
        iarr : np.ndarray
            Uint array with same shape as exclusion extent, representing the 1D
            index values if the geotiff extent was flattened
            (with default flatten order 'C')
        """
        if self._iarr is None:
            N = self.shape[0] * self.shape[1]
            self._iarr = np.arange(N, dtype=np.uint32)
            self._iarr = self._iarr.reshape(self.shape)

        return self._iarr

    @property
    def profile(self):
        """
        GeoTiff profile for exclusions

        Returns
        -------
        profile : dict
        """
        return json.loads(self.h5.global_attrs['profile'])

    @property
    def crs(self):
        """
        GeoTiff projection crs

        Returns
        -------
        str
        """
        return self.profile['crs']

    @property
    def pixel_area(self):
        """Get pixel area in km2 from the transform profile of the excl file.

        Returns
        -------
        area : float
            Exclusion pixel area in km2. Will return None if the
            appropriate transform attribute is not found.
        """

        area = None
        if 'transform' in self.profile:
            transform = self.profile['transform']
            area = np.abs(transform[0] * transform[4])
            area /= 1000 ** 2

        return area

    @property
    def layers(self):
        """
        Available exclusions layers

        Returns
        -------
        layers : list
        """
        layers = self.h5.datasets

        return layers

    @property
    def shape(self):
        """
        Exclusion shape (latitude, longitude)

        Returns
        -------
        shape : tuple
        """
        shape = self.h5.attrs.get('shape', None)
        if shape is None:
            shape = self.h5.shapes[LATITUDE]

        return tuple(shape)

    @property
    def chunks(self):
        """
        Exclusion layers chunks default chunk size

        Returns
        -------
        chunks : tuple | None
            Chunk size of exclusion layers
        """
        chunks = self.h5.attrs.get('chunks', None)
        if chunks is None:
            chunks = self.h5.chunks[LATITUDE]

        return chunks

    @property
    def latitude(self):
        """
        Latitude coordinates array

        Returns
        -------
        ndarray
        """
        return self[LATITUDE]

    @property
    def longitude(self):
        """
        Longitude coordinates array

        Returns
        -------
        ndarray
        """
        return self[LONGITUDE]

    def get_layer_profile(self, layer):
        """
        Get profile for a specific exclusion layer

        Parameters
        ----------
        layer : str
            Layer to get profile for

        Returns
        -------
        profile : dict | None
            GeoTiff profile for single exclusion layer
        """
        profile = self.h5.get_attrs(dset=layer).get('profile', None)
        if profile is not None:
            profile = json.loads(profile)

        return profile

    def get_layer_crs(self, layer):
        """
        Get crs for a specific exclusion layer

        Parameters
        ----------
        layer : str
            Layer to get profile for

        Returns
        -------
        crs : str | None
            GeoTiff projection crs
        """
        profile = self.get_layer_profile(layer)
        if profile is not None:
            crs = profile['crs']
        else:
            crs = None

        return crs

    def get_layer_values(self, layer):
        """
        Get values for given layer in Geotiff format (bands, y, x)

        Parameters
        ----------
        layer : str
            Layer to get values for

        Returns
        -------
        values : ndarray
            GeoTiff values for single exclusion layer
        """
        values = self.h5[layer]

        return values

    def get_layer_description(self, layer):
        """
        Get description for given layer

        Parameters
        ----------
        layer : str
            Layer to get description for

        Returns
        -------
        description : str
            Description of layer
        """
        description = self.h5.get_attrs(dset=layer).get('description', None)

        return description

    def get_nodata_value(self, layer):
        """
        Get the nodata value for a given layer

        Parameters
        ----------
        layer : str
            Layer to get nodata value for

        Returns
        -------
        nodata : int | float | None
            nodata value for layer or None if not found
        """
        profile = self.get_layer_profile(layer)
        nodata = profile.get('nodata', None)

        return nodata

    def _get_latitude(self, *ds_slice):
        """
        Extract latitude coordinates

        Parameters
        ----------
        ds_slice : tuple of int | list | slice
            Pandas slicing describing which sites and columns to extract

        Returns
        -------
        lat : ndarray
            Latitude coordinates
        """
        if LATITUDE not in self.h5:
            msg = ('"latitude" is missing from {}'
                   .format(self.h5_file))
            logger.error(msg)
            raise HandlerKeyError(msg)

        ds_slice = (LATITUDE, ) + ds_slice

        lat = self.h5[ds_slice]

        return lat

    def _get_longitude(self, *ds_slice):
        """
        Extract longitude coordinates

        Parameters
        ----------
        ds_slice : tuple of int | list | slice
            Pandas slicing describing which sites and columns to extract

        Returns
        -------
        lon : ndarray
            Longitude coordinates
        """
        if LONGITUDE not in self.h5:
            msg = ('"longitude" is missing from {}'
                   .format(self.h5_file))
            logger.error(msg)
            raise HandlerKeyError(msg)

        ds_slice = (LONGITUDE, ) + ds_slice

        lon = self.h5[ds_slice]

        return lon

    def _get_layer(self, layer_name, *ds_slice):
        """
        Extract data from given dataset

        Parameters
        ----------
        layer_name : str
            Exclusion layer to extract
        ds_slice : tuple of int | list | slice
            tuple describing slice of layer array to extract

        Returns
        -------
        layer_data : ndarray
            Array of exclusion data
        """
        if layer_name not in self.layers:
            msg = ('{} not in available layers: {}'
                   .format(layer_name, self.layers))
            logger.error(msg)
            raise HandlerKeyError(msg)

        shape = self.h5.get_dset_properties(layer_name)[0]
        if len(shape) == 3:
            ds_slice = (layer_name, 0) + ds_slice
        else:
            ds_slice = (layer_name, ) + ds_slice

        layer_data = self.h5[ds_slice]

        return layer_data



# -*- coding: utf-8 -*-
"""reV-to-SAM generation interface module.

Wraps the NREL-PySAM pvwattsv5, windpower, and tcsmolensalt modules with
additional reV features.
"""

import copy
import logging
import os
from abc import ABC, abstractmethod
from tempfile import TemporaryDirectory
from warnings import warn

import numpy as np
import pandas as pd
import PySAM.Geothermal as PySamGeothermal
import PySAM.LinearFresnelDsgIph as PySamLds
import PySAM.MhkWave as PySamMhkWave
import PySAM.Pvsamv1 as PySamDetailedPv
import PySAM.Pvwattsv5 as PySamPv5
import PySAM.Pvwattsv7 as PySamPv7
import PySAM.Pvwattsv8 as PySamPv8
import PySAM.Swh as PySamSwh
import PySAM.TcsmoltenSalt as PySamCSP
import PySAM.TroughPhysicalProcessHeat as PySamTpph
import PySAM.Windpower as PySamWindPower

from reV.losses import PowerCurveLossesMixin, ScheduledLossesMixin
from reV.SAM.defaults import (
    DefaultGeothermal,
    DefaultLinearFresnelDsgIph,
    DefaultMhkWave,
    DefaultPvSamv1,
    DefaultPvWattsv5,
    DefaultPvWattsv8,
    DefaultSwh,
    DefaultTcsMoltenSalt,
    DefaultTroughPhysicalProcessHeat,
    DefaultWindPower,
)
from reV.SAM.econ import LCOE, SingleOwner
from reV.SAM.SAM import RevPySam
from reV.utilities import ResourceMetaField, SupplyCurveField
from reV.utilities.curtailment import curtail
from reV.utilities.exceptions import (
    InputError,
    SAMExecutionError,
    SAMInputWarning,
)

logger = logging.getLogger(__name__)


class AbstractSamGeneration(RevPySam, ScheduledLossesMixin, ABC):
    """Base class for SAM generation simulations."""

    def __init__(
        self,
        resource,
        meta,
        sam_sys_inputs,
        site_sys_inputs=None,
        output_request=None,
        drop_leap=False,
    ):
        """Initialize a SAM generation object.

        Parameters
        ----------
        resource : pd.DataFrame
            Timeseries solar or wind resource data for a single location with a
            pandas DatetimeIndex.  There must be columns for all the required
            variables to run the respective SAM simulation. Remapping will be
            done to convert typical NSRDB/WTK names into SAM names (e.g. DNI ->
            dn and wind_speed -> windspeed)
        meta : pd.DataFrame | pd.Series
            Meta data corresponding to the resource input for the single
            location. Should include values for latitude, longitude, elevation,
            and timezone.
        sam_sys_inputs : dict
            Site-agnostic SAM system model inputs arguments.
        site_sys_inputs : dict
            Optional set of site-specific SAM system inputs to complement the
            site-agnostic inputs.
        output_request : list
            Requested SAM outputs (e.g., 'cf_mean', 'annual_energy',
            'cf_profile', 'gen_profile', 'energy_yield', 'ppa_price',
            'lcoe_fcr').
        drop_leap : bool
            Drops February 29th from the resource data. If False, December
            31st is dropped from leap years.
        """

        # drop the leap day
        if drop_leap:
            resource = self.drop_leap(resource)

        # make sure timezone and elevation are in the meta data
        meta = self.tz_elev_check(sam_sys_inputs, site_sys_inputs, meta)

        # don't pass resource to base class,
        # set in concrete generation classes instead
        super().__init__(
            meta,
            sam_sys_inputs,
            output_request,
            site_sys_inputs=site_sys_inputs,
        )

        # Set the site number using resource
        if hasattr(resource, "name"):
            self._site = resource.name
        else:
            self._site = None

        # let children pass in None resource
        if resource is not None:
            self.check_resource_data(resource)
            self.set_resource_data(resource, meta)

        self.add_scheduled_losses(resource)

    @classmethod
    def _get_res(cls, res_df, output_request):
        """Get the resource arrays and pass through for output (single site).

        Parameters
        ----------
        res_df : pd.DataFrame
            2D table with resource data.
        output_request : list
            Outputs to retrieve from SAM.

        Returns
        -------
        res_mean : dict | None
            Dictionary object with variables for resource arrays.
        out_req_cleaned : list
            Output request list with the resource request entries removed.
        """

        out_req_cleaned = copy.deepcopy(output_request)
        res_out = None

        res_reqs = []
        ti = res_df.index
        for req in out_req_cleaned:
            if req in res_df:
                res_reqs.append(req)
                if res_out is None:
                    res_out = {}
                res_out[req] = cls.ensure_res_len(res_df[req].values, ti)

        for req in res_reqs:
            out_req_cleaned.remove(req)

        return res_out, out_req_cleaned

    @staticmethod
    def _get_res_mean(resource, res_gid, output_request):
        """Get the resource annual means (single site).

        Parameters
        ----------
        resource : rex.sam_resource.SAMResource
            SAM resource object for WIND resource
        res_gid : int
            Site to extract means for
        output_request : list
            Outputs to retrieve from SAM.

        Returns
        -------
        res_mean : dict | None
            Dictionary object with variables for resource means.
        out_req_nomeans : list
            Output request list with the resource mean entries removed.
        """

        out_req_nomeans = copy.deepcopy(output_request)
        res_mean = None
        idx = resource.sites.index(res_gid)
        irrad_means = (
            "dni_mean",
            "dhi_mean",
            "ghi_mean",
            "clearsky_dni_mean",
            "clearsky_dhi_mean",
            "clearsky_ghi_mean",
        )

        if "ws_mean" in out_req_nomeans:
            out_req_nomeans.remove("ws_mean")
            res_mean = {}
            res_mean["ws_mean"] = resource["mean_windspeed", idx]

        else:
            for var in resource.var_list:
                label_1 = "{}_mean".format(var)
                label_2 = "mean_{}".format(var)
                if label_1 in out_req_nomeans:
                    out_req_nomeans.remove(label_1)
                    if res_mean is None:
                        res_mean = {}
                    res_mean[label_1] = resource[label_2, idx]

                    if label_1 in irrad_means:
                        # convert to kWh/m2/day
                        res_mean[label_1] /= 1000
                        res_mean[label_1] *= 24

        return res_mean, out_req_nomeans

    def check_resource_data(self, resource):
        """Check resource dataframe for NaN values

        Parameters
        ----------
        resource : pd.DataFrame
            Timeseries solar or wind resource data for a single location with a
            pandas DatetimeIndex.  There must be columns for all the required
            variables to run the respective SAM simulation. Remapping will be
            done to convert typical NSRDB/WTK names into SAM names (e.g. DNI ->
            dn and wind_speed -> windspeed)
        """
        if pd.isna(resource).any().any():
            bad_vars = pd.isna(resource).any(axis=0)
            bad_vars = resource.columns[bad_vars].values.tolist()
            msg = "Found NaN values for site {} in variables {}".format(
                self.site, bad_vars
            )
            logger.error(msg)
            raise InputError(msg)

        if len(resource) < 8760:
            msg = (f"Detected resource time series of length {len(resource)}, "
                   "which is less than 8760. This may yield unexpected "
                   "results or fail altogether. If this is not intentional, "
                   "try setting 'time_index_step: 1' in your SAM config or "
                   "double check the resource input you're using.")
            logger.warning(msg)
            warn(msg)

    @abstractmethod
    def set_resource_data(self, resource, meta):
        """Placeholder for resource data setting (nsrdb or wtk)"""

    @staticmethod
    def tz_elev_check(sam_sys_inputs, site_sys_inputs, meta):
        """Check timezone+elevation input and use json config
        timezone+elevation if not in resource meta.

        Parameters
        ----------
        sam_sys_inputs : dict
            Site-agnostic SAM system model inputs arguments.
        site_sys_inputs : dict
            Optional set of site-specific SAM system inputs to complement the
            site-agnostic inputs.
        meta : pd.DataFrame | pd.Series
            Meta data corresponding to the resource input for the single
            location. Should include values for latitude, longitude, elevation,
            and timezone.

        Returns
        -------
        meta : pd.DataFrame | pd.Series
            Dataframe or series for a single site. Will include "timezone"
            and "elevation" from the sam and site system inputs if found.
        """

        if meta is not None:
            axis = 0 if isinstance(meta, pd.core.series.Series) else 1
            meta = meta.rename(
                SupplyCurveField.map_to(ResourceMetaField), axis=axis
            )
            if sam_sys_inputs is not None:
                if ResourceMetaField.ELEVATION in sam_sys_inputs:
                    meta[ResourceMetaField.ELEVATION] = sam_sys_inputs[
                        ResourceMetaField.ELEVATION
                    ]
                if ResourceMetaField.TIMEZONE in sam_sys_inputs:
                    meta[ResourceMetaField.TIMEZONE] = int(
                        sam_sys_inputs[ResourceMetaField.TIMEZONE]
                    )

            # site-specific inputs take priority over generic system inputs
            if site_sys_inputs is not None:
                if ResourceMetaField.ELEVATION in site_sys_inputs:
                    meta[ResourceMetaField.ELEVATION] = site_sys_inputs[
                        ResourceMetaField.ELEVATION
                    ]
                if ResourceMetaField.TIMEZONE in site_sys_inputs:
                    meta[ResourceMetaField.TIMEZONE] = int(
                        site_sys_inputs[ResourceMetaField.TIMEZONE]
                    )

            if ResourceMetaField.TIMEZONE not in meta:
                msg = (
                    "Need timezone input to run SAM gen. Not found in "
                    "resource meta or technology json input config."
                )
                raise SAMExecutionError(msg)

        return meta

    @property
    def has_timezone(self):
        """Returns true if instance has a timezone set"""
        if self._meta is not None and ResourceMetaField.TIMEZONE in self.meta:
            return True

        return False

    def cf_mean(self):
        """Get mean capacity factor (fractional) from SAM.

        Returns
        -------
        output : float
            Mean capacity factor (fractional).
        """
        return self["capacity_factor"] / 100

    def cf_profile(self):
        """Get hourly capacity factor (frac) profile in local timezone.
        See self.outputs attribute for collected output data in UTC.

        Returns
        -------
        cf_profile : np.ndarray
            1D numpy array of capacity factor profile.
            Datatype is float32 and array length is 8760*time_interval.
        """
        return self.gen_profile() / self.sam_sys_inputs["system_capacity"]

    def annual_energy(self):
        """Get annual energy generation value in kWh from SAM.

        Returns
        -------
        output : float
            Annual energy generation (kWh).
        """
        return self["annual_energy"]

    def energy_yield(self):
        """Get annual energy yield value in kwh/kw from SAM.

        Returns
        -------
        output : float
            Annual energy yield (kwh/kw).
        """
        return self["kwh_per_kw"]

    def gen_profile(self):
        """Get power generation profile (local timezone) in kW.
        See self.outputs attribute for collected output data in UTC.

        Returns
        -------
        output : np.ndarray
            1D array of hourly power generation in kW.
            Datatype is float32 and array length is 8760*time_interval.
        """
        return np.array(self["gen"], dtype=np.float32)

    def collect_outputs(self, output_lookup=None):
        """Collect SAM output_request, convert timeseries outputs to UTC, and
        save outputs to self.outputs property.


        Parameters
        ----------
        output_lookup : dict | None
            Lookup dictionary mapping output keys to special output methods.
            None defaults to generation default outputs.
        """

        if output_lookup is None:
            output_lookup = {
                "cf_mean": self.cf_mean,
                "cf_profile": self.cf_profile,
                "annual_energy": self.annual_energy,
                "energy_yield": self.energy_yield,
                "gen_profile": self.gen_profile,
            }

        super().collect_outputs(output_lookup=output_lookup)

    def run_gen_and_econ(self):
        """Run SAM generation with possibility for follow on econ analysis."""

        lcoe_out_reqs = None
        so_out_reqs = None
        lcoe_vars = (
            "lcoe_fcr",
            "fixed_charge_rate",
            "capital_cost",
            "fixed_operating_cost",
            "variable_operating_cost",
        )
        so_vars = (
            "ppa_price",
            "lcoe_real",
            "lcoe_nom",
            "project_return_aftertax_npv",
            "flip_actual_irr",
            "gross_revenue",
        )
        if "lcoe_fcr" in self.output_request:
            lcoe_out_reqs = [r for r in self.output_request if r in lcoe_vars]
            self.output_request = [
                r for r in self.output_request if r not in lcoe_out_reqs
            ]
        elif any(x in self.output_request for x in so_vars):
            so_out_reqs = [r for r in self.output_request if r in so_vars]
            self.output_request = [
                r for r in self.output_request if r not in so_out_reqs
            ]

        # Execute the SAM generation compute module (pvwattsv7, windpower, etc)
        self.run()

        # Execute a follow-on SAM econ compute module
        # (lcoe_fcr, singleowner, etc)
        if lcoe_out_reqs is not None:
            self.sam_sys_inputs["annual_energy"] = self.annual_energy()
            lcoe = LCOE(self.sam_sys_inputs, output_request=lcoe_out_reqs)
            lcoe.assign_inputs()
            lcoe.execute()
            lcoe.collect_outputs()
            self.outputs.update(lcoe.outputs)

        elif so_out_reqs is not None:
            self.sam_sys_inputs["gen"] = self.gen_profile()
            so = SingleOwner(self.sam_sys_inputs, output_request=so_out_reqs)
            so.assign_inputs()
            so.execute()
            so.collect_outputs()
            self.outputs.update(so.outputs)

    def run(self):
        """Run a reV-SAM generation object by assigning inputs, executing the
        SAM simulation, collecting outputs, and converting all arrays to UTC.
        """
        self.assign_inputs()
        self.execute()
        self.collect_outputs()

    @classmethod
    def reV_run(
        cls,
        points_control,
        res_file,
        site_df,
        lr_res_file=None,
        output_request=("cf_mean",),
        drop_leap=False,
        gid_map=None,
        nn_map=None,
        bias_correct=None,
    ):
        """Execute SAM generation based on a reV points control instance.

        Parameters
        ----------
        points_control : config.PointsControl
            PointsControl instance containing project points site and SAM
            config info.
        res_file : str
            Resource file with full path.
        site_df : pd.DataFrame
            Dataframe of site-specific input variables. Row index corresponds
            to site number/gid (via df.loc not df.iloc), column labels are the
            variable keys that will be passed forward as SAM parameters.
        lr_res_file : str | None
            Optional low resolution resource file that will be dynamically
            mapped+interpolated to the nominal-resolution res_file. This
            needs to be of the same format as resource_file, e.g. they both
            need to be handled by the same rex Resource handler such as
            WindResource
        output_request : list | tuple
            Outputs to retrieve from SAM.
        drop_leap : bool
            Drops February 29th from the resource data. If False, December
            31st is dropped from leap years.
        gid_map : None | dict
            Mapping of unique integer generation gids (keys) to single integer
            resource gids (values). This enables the user to input unique
            generation gids in the project points that map to non-unique
            resource gids. This can be None or a pre-extracted dict.
        nn_map : np.ndarray
            Optional 1D array of nearest neighbor mappings associated with the
            res_file to lr_res_file spatial mapping. For details on this
            argument, see the rex.MultiResolutionResource docstring.
        bias_correct : None | pd.DataFrame
            Optional DataFrame or CSV filepath to a wind or solar
            resource bias correction table. This has columns:

                - ``gid``: GID of site (can be index name of dataframe)
                - ``method``: function name from ``rex.bias_correction`` module

            The ``gid`` field should match the true resource ``gid`` regardless
            of the optional ``gid_map`` input. Only ``windspeed`` **or**
            ``GHI`` + ``DNI`` + ``DHI`` are corrected, depending on the
            technology (wind for the former, PV or CSP for the latter). See the
            functions in the ``rex.bias_correction`` module for available
            inputs for ``method``. Any additional kwargs required for the
            requested ``method`` can be input as additional columns in the
            ``bias_correct`` table e.g., for linear bias correction functions
            you can include ``scalar`` and ``adder`` inputs as columns in the
            ``bias_correct`` table on a site-by-site basis. If ``None``, no
            corrections are applied. By default, ``None``.

        Returns
        -------
        out : dict
            Nested dictionaries where the top level key is the site index,
            the second level key is the variable name, second level value is
            the output variable value.
        """
        # initialize output dictionary
        out = {}

        # Get the RevPySam resource object
        resources = RevPySam.get_sam_res(
            res_file,
            points_control.project_points,
            points_control.project_points.tech,
            output_request=output_request,
            gid_map=gid_map,
            lr_res_file=lr_res_file,
            nn_map=nn_map,
            bias_correct=bias_correct,
        )

        # run resource through curtailment filter if applicable
        curtailment = points_control.project_points.curtailment
        if curtailment is not None:
            resources = curtail(
                resources, curtailment, random_seed=curtailment.random_seed
            )

        # iterate through project_points gen_gid values
        for gen_gid in points_control.project_points.sites:
            # Lookup the resource gid if there's a mapping and get the resource
            # data from the SAMResource object using the res_gid.
            res_gid = gen_gid if gid_map is None else gid_map[gen_gid]
            site_res_df, site_meta = resources._get_res_df(res_gid)

            # drop the leap day
            if drop_leap:
                site_res_df = cls.drop_leap(site_res_df)

            _, inputs = points_control.project_points[gen_gid]

            # get resource data pass-throughs and resource means
            res_outs, out_req_cleaned = cls._get_res(
                site_res_df, output_request
            )
            res_mean, out_req_cleaned = cls._get_res_mean(
                resources, res_gid, out_req_cleaned
            )

            # iterate through requested sites.
            sim = cls(
                resource=site_res_df,
                meta=site_meta,
                sam_sys_inputs=inputs,
                output_request=out_req_cleaned,
                site_sys_inputs=dict(site_df.loc[gen_gid, :]),
            )
            sim.run_gen_and_econ()

            # collect outputs to dictout
            out[gen_gid] = sim.outputs

            if res_outs is not None:
                out[gen_gid].update(res_outs)

            if res_mean is not None:
                out[gen_gid].update(res_mean)

        return out


class AbstractSamGenerationFromWeatherFile(AbstractSamGeneration, ABC):
    """Base class for running sam generation with a weather file on disk."""

    WF_META_DROP_COLS = {
        ResourceMetaField.LATITUDE,
        ResourceMetaField.LONGITUDE,
        ResourceMetaField.ELEVATION,
        ResourceMetaField.TIMEZONE,
        ResourceMetaField.COUNTRY,
        ResourceMetaField.STATE,
        ResourceMetaField.COUNTY,
        "urban",
        "population",
        "landcover",
    }

    @property
    @abstractmethod
    def PYSAM_WEATHER_TAG(self):
        """Name of the weather file input used by SAM generation module."""
        raise NotImplementedError

    def set_resource_data(self, resource, meta):
        """Generate the weather file and set the path as an input.

        Some PySAM models require a data file, not raw data. This method
        generates the weather data, writes it to a file on disk, and
        then sets the file as an input to the generation module. The
        function
        :meth:`~AbstractSamGenerationFromWeatherFile.run_gen_and_econ`
        deletes the file on disk after a run is complete.

        Parameters
        ----------
        resource : pd.DataFrame
            Time series resource data for a single location with a
            pandas DatetimeIndex. There must be columns for all the
            required variables to run the respective SAM simulation.
            Remapping will be done to convert typical NSRDB/WTK names
            into SAM names (e.g. DNI -> dn and wind_speed -> windspeed).
        meta : pd.Series
            Meta data corresponding to the resource input for the single
            location. Should include values for latitude, longitude,
            elevation, and timezone.
        """
        meta = self._parse_meta(meta)
        self.time_interval = self.get_time_interval(resource.index.values)
        pysam_w_fname = self._create_pysam_wfile(resource, meta)
        self[self.PYSAM_WEATHER_TAG] = pysam_w_fname

    def _create_pysam_wfile(self, resource, meta):
        """Create PySAM weather input file.

        Parameters
        ----------
        resource : pd.DataFrame
            Time series resource data for a single location with a
            pandas DatetimeIndex. There must be columns for all the
            required variables to run the respective SAM simulation.
            Remapping will be done to convert typical NSRDB/WTK names
            into SAM names (e.g. DNI -> dn and wind_speed -> windspeed).
        meta : pd.Series
            Meta data corresponding to the resource input for the single
            location. Should include values for latitude, longitude,
            elevation, and timezone.

        Returns
        -------
        fname : str
            Name of weather csv file.

        Notes
        -----
        PySAM will not accept data on Feb 29th. For leap years,
        December 31st is dropped and time steps are shifted to relabel
        Feb 29th as March 1st, March 1st as March 2nd, etc.
        """
        # pylint: disable=attribute-defined-outside-init,consider-using-with
        self._temp_dir = TemporaryDirectory()
        fname = os.path.join(self._temp_dir.name, "weather.csv")
        logger.debug("Creating PySAM weather data file: {}".format(fname))

        # ------- Process metadata
        m = pd.DataFrame(meta).T
        timezone = m[ResourceMetaField.TIMEZONE]
        m["Source"] = "NSRDB"
        m["Location ID"] = meta.name
        m["City"] = "-"
        m["State"] = m["state"].apply(lambda x: "-" if x == "None" else x)
        m["Country"] = m["country"].apply(lambda x: "-" if x == "None" else x)
        m["Latitude"] = m[ResourceMetaField.LATITUDE]
        m["Longitude"] = m[ResourceMetaField.LONGITUDE]
        m["Time Zone"] = timezone
        m["Elevation"] = m[ResourceMetaField.ELEVATION]
        m["Local Time Zone"] = timezone
        m["Dew Point Units"] = "c"
        m["DHI Units"] = "w/m2"
        m["DNI Units"] = "w/m2"
        m["Temperature Units"] = "c"
        m["Pressure Units"] = "mbar"
        m["Wind Speed"] = "m/s"
        keep_cols = [c for c in m.columns if c not in self.WF_META_DROP_COLS]
        m[keep_cols].to_csv(fname, index=False, mode="w")

        # --------- Process data
        var_map = {
            "dni": "DNI",
            "dhi": "DHI",
            "wind_speed": "Wind Speed",
            "air_temperature": "Temperature",
            "dew_point": "Dew Point",
            "surface_pressure": "Pressure",
        }
        resource = resource.rename(mapper=var_map, axis="columns")

        time_index = resource.index
        # Adjust from UTC to local time
        local = np.roll(
            resource.values, int(timezone * self.time_interval), axis=0
        )
        resource = pd.DataFrame(
            local, columns=resource.columns, index=time_index
        )
        mask = (time_index.month == 2) & (time_index.day == 29)
        time_index = time_index[~mask]

        df = pd.DataFrame(index=time_index)
        df["Year"] = time_index.year
        df["Month"] = time_index.month
        df["Day"] = time_index.day
        df["Hour"] = time_index.hour
        df["Minute"] = time_index.minute
        df = df.join(resource.loc[~mask])

        df.to_csv(fname, index=False, mode="a")

        return fname

    def run_gen_and_econ(self):
        """Run SAM generation and possibility follow-on econ analysis."""
        try:
            super().run_gen_and_econ()
        finally:
            temp_dir = getattr(self, "_temp_dir", None)
            if temp_dir is not None:
                temp_dir.cleanup()


class AbstractSamSolar(AbstractSamGeneration, ABC):
    """Base Class for Solar generation from SAM"""

    @staticmethod
    def agg_albedo(time_index, albedo):
        """Aggregate a timeseries of albedo data to monthly values w len 12 as
        required by pysam Pvsamv1

        Tech spec from pysam docs:
        https://nrel-pysam.readthedocs.io/en/master/modules/Pvsamv1.html
        #PySAM.Pvsamv1.Pvsamv1.SolarResource.albedo

        Parameters
        ----------
        time_index : pd.DatetimeIndex
            Timeseries solar resource datetimeindex
        albedo : list
            Timeseries Albedo data to be aggregated. Should be 0-1 and likely
            hourly or less.

        Returns
        -------
        monthly_albedo : list
            1D list of monthly albedo values with length 12
        """
        monthly_albedo = np.zeros(12).tolist()
        albedo = np.array(albedo)
        for month in range(1, 13):
            m = np.where(time_index.month == month)[0]
            monthly_albedo[int(month - 1)] = albedo[m].mean()

        return monthly_albedo

    def set_resource_data(self, resource, meta):
        """Set NSRDB resource data arrays.

        Parameters
        ----------
        resource : pd.DataFrame
            Timeseries solar or wind resource data for a single location with a
            pandas DatetimeIndex.  There must be columns for all the required
            variables to run the respective SAM simulation. Remapping will be
            done to convert typical NSRDB/WTK names into SAM names (e.g. DNI ->
            dn and wind_speed -> windspeed)
        meta : pd.Series
            Meta data corresponding to the resource input for the single
            location. Should include values for latitude, longitude, elevation,
            and timezone.
        """

        meta = self._parse_meta(meta)
        time_index = resource.index
        self.time_interval = self.get_time_interval(resource.index.values)

        # map resource data names to SAM required data names
        var_map = {
            "dni": "dn",
            "dhi": "df",
            "ghi": "gh",
            "clearskydni": "dn",
            "clearskydhi": "df",
            "clearskyghi": "gh",
            "windspeed": "wspd",
            "airtemperature": "tdry",
            "temperature": "tdry",
            "temp": "tdry",
            "dewpoint": "tdew",
            "surfacepressure": "pres",
            "pressure": "pres",
            "surfacealbedo": "albedo",
        }
        lower_case = {
            k: k.lower().replace(" ", "").replace("_", "")
            for k in resource.columns
        }
        irrad_vars = ["dn", "df", "gh"]

        resource = resource.rename(mapper=lower_case, axis="columns")
        resource = resource.rename(mapper=var_map, axis="columns")
        time_index = resource.index
        resource = {
            k: np.array(v)
            for (k, v) in resource.to_dict(orient="list").items()
        }

        # set resource variables
        for var, arr in resource.items():
            if var != "time_index":
                # ensure that resource array length is multiple of 8760
                arr = self.ensure_res_len(arr, time_index)
                n_roll = int(
                    self._meta[ResourceMetaField.TIMEZONE] * self.time_interval
                )
                arr = np.roll(arr, n_roll)

                if var in irrad_vars and np.min(arr) < 0:
                    warn(
                        'Solar irradiance variable "{}" has a minimum '
                        "value of {}. Truncating to zero.".format(
                            var, np.min(arr)
                        ),
                        SAMInputWarning,
                    )
                    arr = np.where(arr < 0, 0, arr)

                resource[var] = arr.tolist()

        resource["lat"] = meta[ResourceMetaField.LATITUDE]
        resource["lon"] = meta[ResourceMetaField.LONGITUDE]
        resource["tz"] = meta[ResourceMetaField.TIMEZONE]

        resource["elev"] = meta.get(ResourceMetaField.ELEVATION, 0.0)

        time_index = self.ensure_res_len(time_index, time_index)
        resource["minute"] = time_index.minute
        resource["hour"] = time_index.hour
        resource["month"] = time_index.month
        resource["year"] = time_index.year
        resource["day"] = time_index.day

        if "albedo" in resource:
            self["albedo"] = self.agg_albedo(
                time_index, resource.pop("albedo")
            )

        self["solar_resource_data"] = resource


class AbstractSamPv(AbstractSamSolar, ABC):
    """Photovoltaic (PV) generation with either pvwatts of detailed pv."""

    # set these class attrs in concrete subclasses
    MODULE = None
    PYSAM = None

    # pylint: disable=line-too-long
    def __init__(
        self,
        resource,
        meta,
        sam_sys_inputs,
        site_sys_inputs=None,
        output_request=None,
        drop_leap=False,
    ):
        """Initialize a SAM solar object.

        See the PySAM :py:class:`~PySAM.Pvwattsv8.Pvwattsv8` (or older
        version model) or :py:class:`~PySAM.Pvsamv1.Pvsamv1` documentation for
        the configuration keys required in the `sam_sys_inputs` config for the
        respective models. Some notable keys include the following to enable a
        lifetime simulation (non-exhaustive):

            - ``system_use_lifetime_output`` : Integer flag indicating whether
              or not to run a full lifetime model (0 for off, 1 for on). If
              running a lifetime model, the resource file will be repeated
              for the number of years specified as the lifetime of the
              plant and a performance degradation term will be used to
              simulate reduced performance over time.
            - ``analysis_period`` : Integer representing the number of years
              to include in the lifetime of the model generator. Required if
              ``system_use_lifetime_output`` is set to 1.
            - ``dc_degradation`` : List of percentage values representing the
              annual DC degradation of capacity factors. Maybe a single value
              that will be compound each year or a vector of yearly rates.
              Required if ``system_use_lifetime_output`` is set to 1.

        You may also include the following ``reV``-specific keys:

            - ``reV_outages`` : Specification for ``reV``-scheduled
              stochastic outage losses. For example::

                    outage_info = [
                        {
                            'count': 6,
                            'duration': 24,
                            'percentage_of_capacity_lost': 100,
                            'allowed_months': ['January', 'March'],
                            'allow_outage_overlap': True
                        },
                        {
                            'count': 10,
                            'duration': 1,
                            'percentage_of_capacity_lost': 10,
                            'allowed_months': ['January'],
                            'allow_outage_overlap': False
                        },
                        ...
                    ]

              See the description of
              :meth:`~reV.losses.scheduled.ScheduledLossesMixin.add_scheduled_losses`
              or the
              `reV losses demo notebook <https://tinyurl.com/4d7uutt3/>`_
              for detailed instructions on how to specify this input.
            - ``reV_outages_seed`` : Integer value used to seed the RNG
              used to compute stochastic outage losses.
            - ``time_index_step`` : Integer representing the step size
              used to sample the ``time_index`` in the resource data.
              This can be used to reduce temporal resolution (i.e. for
              30 minute NSRDB input data, ``time_index_step=1`` yields
              the full 30 minute time series as output, while
              ``time_index_step=2`` yields hourly output, and so forth).

              .. Note:: The reduced data shape (i.e. after applying a
                        step size of `time_index_step`) must still be an
                        integer multiple of 8760, or the execution will
                        fail.

            - ``clearsky`` : Boolean flag value indicating wether
              computation should use clearsky resource data to compute
              generation data.

        Parameters
        ----------
        resource : pd.DataFrame
            Timeseries solar or wind resource data for a single location with a
            pandas DatetimeIndex.  There must be columns for all the required
            variables to run the respective SAM simulation. Remapping will be
            done to convert typical NSRDB/WTK names into SAM names (e.g. DNI ->
            dn and wind_speed -> windspeed)
        meta : pd.DataFrame | pd.Series
            Meta data corresponding to the resource input for the single
            location. Should include values for latitude, longitude, elevation,
            and timezone.
        sam_sys_inputs : dict
            Site-agnostic SAM system model inputs arguments.
        site_sys_inputs : dict
            Optional set of site-specific SAM system inputs to complement the
            site-agnostic inputs.
        output_request : list
            Requested SAM outputs (e.g., 'cf_mean', 'annual_energy',
            'cf_profile', 'gen_profile', 'energy_yield', 'ppa_price',
            'lcoe_fcr').
        drop_leap : bool
            Drops February 29th from the resource data. If False, December
            31st is dropped from leap years.
        """

        # need to check tilt=lat and azimuth for pv systems
        meta = self._parse_meta(meta)
        sam_sys_inputs = self.set_latitude_tilt_az(sam_sys_inputs, meta)

        super().__init__(
            resource,
            meta,
            sam_sys_inputs,
            site_sys_inputs=site_sys_inputs,
            output_request=output_request,
            drop_leap=drop_leap,
        )

    def set_resource_data(self, resource, meta):
        """Set NSRDB resource data arrays.

        Parameters
        ----------
        resource : pd.DataFrame
            Timeseries solar or wind resource data for a single location with a
            pandas DatetimeIndex.  There must be columns for all the required
            variables to run the respective SAM simulation. Remapping will be
            done to convert typical NSRDB/WTK names into SAM names (e.g. DNI ->
            dn and wind_speed -> windspeed)
        meta : pd.Series
            Meta data corresponding to the resource input for the single
            location. Should include values for latitude, longitude, elevation,
            and timezone.

        Raises
        ------
        ValueError : If lat/lon outside of -90 to 90 and -180 to 180,
                     respectively.

        """
        bad_location_input = (
            (meta[ResourceMetaField.LATITUDE] < -90)
            | (meta[ResourceMetaField.LATITUDE] > 90)
            | (meta[ResourceMetaField.LONGITUDE] < -180)
            | (meta[ResourceMetaField.LONGITUDE] > 180)
        )
        if bad_location_input.any():
            raise ValueError(
                "Detected latitude/longitude values outside of "
                "the range -90 to 90 and -180 to 180, "
                "respectively. Please ensure input resource data"
                "locations conform to these ranges. "
            )
        return super().set_resource_data(resource, meta)

    @staticmethod
    def set_latitude_tilt_az(sam_sys_inputs, meta):
        """Check if tilt is specified as latitude and set tilt=lat, az=180 or 0

        Parameters
        ----------
        sam_sys_inputs : dict
            Site-agnostic SAM system model inputs arguments.
        meta : pd.Series
            Meta data corresponding to the resource input for the single
            location. Should include values for latitude, longitude, elevation,
            and timezone.

        Returns
        -------
        sam_sys_inputs : dict
            Site-agnostic SAM system model inputs arguments.
            If for a pv simulation the "tilt" parameter was originally not
            present or set to 'lat' or MetaKeyName.LATITUDE, the tilt will be
            set to the absolute value of the latitude found in meta and the
            azimuth will be 180 if lat>0, 0 if lat<0.
        """

        set_tilt = False
        if sam_sys_inputs is not None and meta is not None:
            if "tilt" not in sam_sys_inputs:
                warn(
                    "No tilt specified, setting at latitude.", SAMInputWarning
                )
                set_tilt = True
            elif (
                sam_sys_inputs["tilt"] == "lat"
                or sam_sys_inputs["tilt"] == ResourceMetaField.LATITUDE
            ) or (
                sam_sys_inputs["tilt"] == "lat"
                or sam_sys_inputs["tilt"] == ResourceMetaField.LATITUDE
            ):
                set_tilt = True

        if set_tilt:
            # set tilt to abs(latitude)
            sam_sys_inputs["tilt"] = np.abs(meta[ResourceMetaField.LATITUDE])
            if meta[ResourceMetaField.LATITUDE] > 0:
                # above the equator, az = 180
                sam_sys_inputs["azimuth"] = 180
            else:
                # below the equator, az = 0
                sam_sys_inputs["azimuth"] = 0

            logger.debug(
                'Tilt specified at "latitude", setting tilt to: {}, '
                "azimuth to: {}".format(
                    sam_sys_inputs["tilt"], sam_sys_inputs["azimuth"]
                )
            )
        return sam_sys_inputs

    def system_capacity_ac(self):
        """Get AC system capacity from SAM inputs.

        NOTE: AC nameplate = DC nameplate / ILR

        Returns
        -------
        cf_profile : float
            AC nameplate = DC nameplate / ILR
        """
        return (
            self.sam_sys_inputs["system_capacity"]
            / self.sam_sys_inputs["dc_ac_ratio"]
        )

    def cf_mean(self):
        """Get mean capacity factor (fractional) from SAM.

        NOTE: PV capacity factor is the AC power production / the DC nameplate

        Returns
        -------
        output : float
            Mean capacity factor (fractional).
            PV CF is calculated as AC power / DC nameplate.
        """
        return self["capacity_factor"] / 100

    def cf_mean_ac(self):
        """Get mean AC capacity factor (fractional) from SAM.

        NOTE: This value only available in PVWattsV8 and up.

        Returns
        -------
        output : float
            Mean AC capacity factor (fractional).
            PV AC CF is calculated as AC power / AC nameplate.
        """
        return self["capacity_factor_ac"] / 100

    def cf_profile(self):
        """Get hourly capacity factor (frac) profile in local timezone.
        See self.outputs attribute for collected output data in UTC.

        NOTE: PV capacity factor is the AC power production / the DC nameplate

        Returns
        -------
        cf_profile : np.ndarray
            1D numpy array of capacity factor profile.
            Datatype is float32 and array length is 8760*time_interval.
            PV CF is calculated as AC power / DC nameplate.
        """
        return self.gen_profile() / self.sam_sys_inputs["system_capacity"]

    def cf_profile_ac(self):
        """Get hourly AC capacity factor (frac) profile in local timezone.
        See self.outputs attribute for collected output data in UTC.

        NOTE: PV AC capacity factor is the AC power production / the AC
        nameplate. AC nameplate = DC nameplate / ILR

        Returns
        -------
        cf_profile : np.ndarray
            1D numpy array of capacity factor profile.
            Datatype is float32 and array length is 8760*time_interval.
            PV AC CF is calculated as AC power / AC nameplate.
        """
        return self.gen_profile() / self.system_capacity_ac()

    def gen_profile(self):
        """Get AC inverter power generation profile (local timezone) in kW.
        This is an alias of the "ac" SAM output variable if PySAM version>=3.
        See self.outputs attribute for collected output data in UTC.

        Returns
        -------
        output : np.ndarray
            1D array of AC inverter power generation in kW.
            Datatype is float32 and array length is 8760*time_interval.
        """
        return np.array(self["gen"], dtype=np.float32)

    def ac(self):
        """Get AC inverter power generation profile (local timezone) in kW.
        See self.outputs attribute for collected output data in UTC.

        Returns
        -------
        output : np.ndarray
            1D array of AC inverter power generation in kW.
            Datatype is float32 and array length is 8760*time_interval.
        """
        return np.array(self["ac"], dtype=np.float32) / 1000

    def dc(self):
        """
        Get DC array power generation profile (local timezone) in kW.
        See self.outputs attribute for collected output data in UTC.

        Returns
        -------
        output : np.ndarray
            1D array of DC array power generation in kW.
            Datatype is float32 and array length is 8760*time_interval.
        """
        return np.array(self["dc"], dtype=np.float32) / 1000

    def clipped_power(self):
        """
        Get the clipped DC power generated behind the inverter
        (local timezone) in kW.
        See self.outputs attribute for collected output data in UTC.

        Returns
        -------
        clipped : np.ndarray
            1D array of clipped DC power in kW.
            Datatype is float32 and array length is 8760*time_interval.
        """
        ac = self.ac()
        dc = self.dc()

        return np.where(ac < ac.max(), 0, dc - ac)

    @staticmethod
    @abstractmethod
    def default():
        """Get the executed default pysam object."""

    def collect_outputs(self, output_lookup=None):
        """Collect SAM output_request, convert timeseries outputs to UTC, and
        save outputs to self.outputs property.

        Parameters
        ----------
        output_lookup : dict | None
            Lookup dictionary mapping output keys to special output methods.
            None defaults to generation default outputs.
        """

        if output_lookup is None:
            output_lookup = {
                "cf_mean": self.cf_mean,
                "cf_mean_ac": self.cf_mean_ac,
                "cf_profile": self.cf_profile,
                "cf_profile_ac": self.cf_profile_ac,
                "annual_energy": self.annual_energy,
                "energy_yield": self.energy_yield,
                "gen_profile": self.gen_profile,
                "ac": self.ac,
                "dc": self.dc,
                "clipped_power": self.clipped_power,
                "system_capacity_ac": self.system_capacity_ac,
            }

        super().collect_outputs(output_lookup=output_lookup)


class PvWattsv5(AbstractSamPv):
    """Photovoltaic (PV) generation with pvwattsv5."""

    MODULE = "pvwattsv5"
    PYSAM = PySamPv5

    @staticmethod
    def default():
        """Get the executed default pysam PVWATTSV5 object.

        Returns
        -------
        PySAM.Pvwattsv5
        """
        return DefaultPvWattsv5.default()


class PvWattsv7(AbstractSamPv):
    """Photovoltaic (PV) generation with pvwattsv7."""

    MODULE = "pvwattsv7"
    PYSAM = PySamPv7

    @staticmethod
    def default():
        """Get the executed default pysam PVWATTSV7 object.

        Returns
        -------
        PySAM.Pvwattsv7
        """
        raise NotImplementedError("Pvwattsv7 default file no longer exists!")


class PvWattsv8(AbstractSamPv):
    """Photovoltaic (PV) generation with pvwattsv8."""

    MODULE = "pvwattsv8"
    PYSAM = PySamPv8

    @staticmethod
    def default():
        """Get the executed default pysam PVWATTSV8 object.

        Returns
        -------
        PySAM.Pvwattsv8
        """
        return DefaultPvWattsv8.default()


class PvSamv1(AbstractSamPv):
    """Detailed PV model"""

    MODULE = "Pvsamv1"
    PYSAM = PySamDetailedPv

    def ac(self):
        """Get AC inverter power generation profile (local timezone) in kW.
        See self.outputs attribute for collected output data in UTC.

        Returns
        -------
        output : np.ndarray
            1D array of AC inverter power generation in kW.
            Datatype is float32 and array length is 8760*time_interval.
        """
        return np.array(self["gen"], dtype=np.float32)

    def dc(self):
        """
        Get DC array power generation profile (local timezone) in kW.
        See self.outputs attribute for collected output data in UTC.

        Returns
        -------
        output : np.ndarray
            1D array of DC array power generation in kW.
            Datatype is float32 and array length is 8760*time_interval.
        """
        return np.array(self["dc_net"], dtype=np.float32)

    @staticmethod
    def default():
        """Get the executed default pysam Pvsamv1 object.

        Returns
        -------
        PySAM.Pvsamv1
        """
        return DefaultPvSamv1.default()


class TcsMoltenSalt(AbstractSamSolar):
    """Concentrated Solar Power (CSP) generation with tower molten salt"""

    MODULE = "tcsmolten_salt"
    PYSAM = PySamCSP

    def cf_profile(self):
        """Get absolute value hourly capacity factor (frac) profile in
        local timezone.
        See self.outputs attribute for collected output data in UTC.

        Returns
        -------
        cf_profile : np.ndarray
            1D numpy array of capacity factor profile.
            Datatype is float32 and array length is 8760*time_interval.
        """
        x = np.abs(self.gen_profile() / self.sam_sys_inputs["system_capacity"])
        return x

    @staticmethod
    def default():
        """Get the executed default pysam CSP object.

        Returns
        -------
        PySAM.TcsmoltenSalt
        """
        return DefaultTcsMoltenSalt.default()


class SolarWaterHeat(AbstractSamGenerationFromWeatherFile):
    """
    Solar Water Heater generation
    """

    MODULE = "solarwaterheat"
    PYSAM = PySamSwh
    PYSAM_WEATHER_TAG = "solar_resource_file"

    @staticmethod
    def default():
        """Get the executed default pysam swh object.

        Returns
        -------
        PySAM.Swh
        """
        return DefaultSwh.default()


class LinearDirectSteam(AbstractSamGenerationFromWeatherFile):
    """
    Process heat linear Fresnel direct steam generation
    """

    MODULE = "lineardirectsteam"
    PYSAM = PySamLds
    PYSAM_WEATHER_TAG = "file_name"

    def cf_mean(self):
        """Calculate mean capacity factor (fractional) from SAM.

        Returns
        -------
        output : float
            Mean capacity factor (fractional).
        """
        net_power = (
            self["annual_field_energy"] - self["annual_thermal_consumption"]
        )  # kW-hr
        # q_pb_des is in MW, convert to kW-hr
        name_plate = self["q_pb_des"] * 8760 * 1000

        return net_power / name_plate

    @staticmethod
    def default():
        """Get the executed default pysam linear Fresnel object.

        Returns
        -------
        PySAM.LinearFresnelDsgIph
        """
        return DefaultLinearFresnelDsgIph.default()


class TroughPhysicalHeat(AbstractSamGenerationFromWeatherFile):
    """
    Trough Physical Process Heat generation
    """

    MODULE = "troughphysicalheat"
    PYSAM = PySamTpph
    PYSAM_WEATHER_TAG = "file_name"

    def cf_mean(self):
        """Calculate mean capacity factor (fractional) from SAM.

        Returns
        -------
        output : float
            Mean capacity factor (fractional).
        """
        net_power = (
            self["annual_gross_energy"] - self["annual_thermal_consumption"]
        )  # kW-hr
        # q_pb_des is in MW, convert to kW-hr
        name_plate = self["q_pb_design"] * 8760 * 1000

        return net_power / name_plate

    @staticmethod
    def default():
        """Get the executed default pysam trough object.

        Returns
        -------
        PySAM.TroughPhysicalProcessHeat
        """
        return DefaultTroughPhysicalProcessHeat.default()


# pylint: disable=line-too-long
class Geothermal(AbstractSamGenerationFromWeatherFile):
    """reV-SAM geothermal generation.

    As of 12/20/2022, the resource potential input in SAM is only used
    to calculate the number of well replacements during the lifetime of
    a geothermal plant. It was decided that reV would not model well
    replacements. Therefore, reV sets the resource potential to match
    (or be just above) the gross potential so that SAM does not throw
    any errors.

    Also as of 12/20/2022, the SAM GETEM module requires a weather file,
    but does not actually require any weather data to run. Therefore,
    reV currently generates an empty weather file to pass to SAM. This
    behavior can be easily updated in the future should the SAM GETEM
    module start using weather data.

    See the PySAM :py:class:`~PySAM.Geothermal.Geothermal` documentation
    for the configuration keys required in the `sam_sys_inputs` config.
    Some notable keys include (non-exhaustive):

        - ``resource_type`` : Integer flag representing either
          Hydrothermal (0) or EGS (1) resource. Only values of 0 or 1
          allowed.
        - ``resource_potential`` : Total resource potential at location
          (in MW).

          .. Important:: ``reV`` automatically sets the resource
             potential to match the gross potential (see documentation
             above), so this key should be left out of the config (it
             will be overridden in any case).

        - ``resource_temp`` : Temperature of resource (in C).

          .. Important:: This value is set by ``reV`` based on the
             user's geothermal resource data input. To override this
             behavior, users *may* specify their own ``resource_temp``
             value (either a single value for all sites in the SAM
             geothermal config or a  site-dependent value in the project
             points CSV). In this case, the resource temperature from
             the input data will be ignored completely, and the
             temperature at each location will be determined solely from
             this input.

        - ``resource_depth`` : Depth to geothermal resource (in m).
        - ``analysis_type`` : Integer flag representing the plant
          configuration. If the ``nameplate`` input is to be used to
          specify the plant capacity, then this flag should be set to 0
          (this is the default ``reV`` assumption). Otherwise, if the
          ``num_wells`` input is to be used to specify the plant site,
          then this flag should be set to 1. Only values of 0 or 1
          allowed.

        - ``nameplate`` : Geothermal plant size (in kW). Only affects
          the output if ``analysis_type=0``.

          .. Important:: Unlike wind or solar, ``reV`` geothermal
             dynamically sets the size of a geothermal plant. In
             particular, the plant capacity is set to match the resource
             potential (obtained from the input data) for each site. For
             this to work, users **must** leave out the ``nameplate``
             key from the SAM config.

             Alternatively, users *may* specify their own ``nameplate``
             capacity value (either a single value for all sites in the
             SAM geothermal config or a site-dependent value in the
             project points CSV). In this case, the resource potential
             from the input data will be ignored completely, and the
             capacity at each location will be determined solely from
             this input.

        - ``num_wells`` : Number of wells at each plant. This value is
          used to determined plant capacity if ``analysis_type=1``.
          Otherwise this input has no effect.
        - ``num_wells_getem`` : Number of wells assumed at each plant
          for power block calculations. Only affects power block outputs
          if ``analysis_type=0`` (otherwise the ``num_wells`` input is
          used in power block calculations).

          .. Note:: ``reV`` does not currently adjust this value based
             on the resource input (as it probably should). If any
             power block outputs are required in the future, there may
             need to be extra development to set this value based on
             the dynamically calculated plant size.

        - ``conversion_type`` : Integer flag representing the conversion
          plant type. Either Binary (0) or Flash (1). Only values of 0
          or 1 allowed.
        - ``geotherm.cost.inj_prod_well_ratio`` : Fraction representing
          the injection to production well ratio (0-1). SAM GUI defaults
          to 0.5 for this value, but it is recommended to set this to
          the GETEM default of 0.75.


    You may also include the following ``reV``-specific keys:

        - ``num_confirmation_wells`` : Number of confirmation wells that
          can also be used as production wells. This number is used to
          determined to total number of wells required at each plant,
          and therefore the total drilling costs. This value defaults to
          2 (to match the SAM GUI as of 8/1/2023). However, the default
          value can lead to negative costs if the plant size is small
          (e.g. only 1 production well is needed, so the costs equal
          -1 * ``drill_cost_per_well``). This is a limitation of the
          SAM calculations (as of 8/1/2023), and it is therefore useful
          to set ``num_confirmation_wells=0`` when performing ``reV``
          runs for small plant sizes.
        - ``capital_cost_per_kw`` : Capital cost values in $/kW. If
          this value is specified in the config, reV calculates and
          overrides the total ``capital_cost`` value based on the
          geothermal plant size (capacity) at each location.
        - ``fixed_operating_cost`` : Fixed operating cost values in
          $/kW. If this value is specified in the config, reV calculates
          and overrides the total ``fixed_operating_cost`` value based
          on the geothermal plant size (capacity) at each location.
        - ``drill_cost_per_well`` : Drilling cost per well, in $. If
          this value is specified in the config, reV calculates the
          total drilling costs based on the number of wells that need to
          be drilled at each location. The drilling costs are added to
          the total ``capital_cost`` at each location.
        - ``reV_outages`` : Specification for ``reV``-scheduled
          stochastic outage losses. For example::

                outage_info = [
                    {
                        'count': 6,
                        'duration': 24,
                        'percentage_of_capacity_lost': 100,
                        'allowed_months': ['January', 'March'],
                        'allow_outage_overlap': True
                    },
                    {
                        'count': 10,
                        'duration': 1,
                        'percentage_of_capacity_lost': 10,
                        'allowed_months': ['January'],
                        'allow_outage_overlap': False
                    },
                    ...
                ]

          See the description of
          :meth:`~reV.losses.scheduled.ScheduledLossesMixin.add_scheduled_losses`
          or the
          `reV losses demo notebook <https://tinyurl.com/4d7uutt3/>`_
          for detailed instructions on how to specify this input.
        - ``reV_outages_seed`` : Integer value used to seed the RNG
          used to compute stochastic outage losses.
        - ``time_index_step`` : Integer representing the step size
          used to sample the ``time_index`` in the resource data.
          This can be used to reduce temporal resolution (i.e. for
          30 minute NSRDB input data, ``time_index_step=1`` yields
          the full 30 minute time series as output, while
          ``time_index_step=2`` yields hourly output, and so forth).

    """

    MODULE = "geothermal"
    PYSAM = PySamGeothermal
    PYSAM_WEATHER_TAG = "file_name"
    _RESOURCE_POTENTIAL_MULT = 1.001
    _DEFAULT_NUM_CONFIRMATION_WELLS = 2  # SAM GUI default as of 5/26/23

    @staticmethod
    def default():
        """Get the executed default PySAM Geothermal object.

        Returns
        -------
        PySAM.Geothermal
        """
        return DefaultGeothermal.default()

    def cf_profile(self):
        """Get hourly capacity factor (frac) profile in local timezone.
        See self.outputs attribute for collected output data in UTC.

        Returns
        -------
        cf_profile : np.ndarray
            1D numpy array of capacity factor profile.
            Datatype is float32 and array length is 8760*time_interval.
        """
        return self.gen_profile() / self.sam_sys_inputs["nameplate"]

    def assign_inputs(self):
        """Assign the self.sam_sys_inputs attribute to the PySAM object."""
        if self.sam_sys_inputs.get("ui_calculations_only"):
            msg = (
                "reV requires model run - cannot set "
                '"ui_calculations_only" to `True` (1). Automatically '
                "setting to `False` (0)!"
            )
            logger.warning(msg)
            warn(msg)
            self.sam_sys_inputs["ui_calculations_only"] = 0
        super().assign_inputs()

    def set_resource_data(self, resource, meta):
        """Generate the weather file and set the path as an input.

        The Geothermal PySAM model requires a data file, not raw data.
        This method generates the weather data, writes it to a file on
        disk, and then sets the file as an input to the Geothermal
        generation module. The function
        :meth:`~AbstractSamGenerationFromWeatherFile.run_gen_and_econ`
        deletes the file on disk after a run is complete.

        Parameters
        ----------
        resource : pd.DataFrame
            Time series resource data for a single location with a
            pandas DatetimeIndex. There must be columns for all the
            required variables to run the respective SAM simulation.
        meta : pd.Series
            Meta data corresponding to the resource input for the single
            location. Should include values for latitude, longitude,
            elevation, and timezone.
        """
        super().set_resource_data(resource, meta)
        self._set_resource_temperature(resource)
        self._set_egs_plant_design_temperature()
        self._set_nameplate_to_match_resource_potential(resource)
        self._set_resource_potential_to_match_gross_output()
        self._set_costs()

    def _set_resource_temperature(self, resource):
        """Set resource temp from data if user did not specify it."""

        if "resource_temp" in self.sam_sys_inputs:
            logger.debug(
                "Found 'resource_temp' value in SAM config: {:.2f}".format(
                    self.sam_sys_inputs["resource_temp"]
                )
            )
            return

        val = set(resource["temperature"].unique())
        logger.debug(
            "Found {} value(s) for 'temperature' in resource data".format(
                len(val)
            )
        )
        if len(val) > 1:
            msg = (
                "Found multiple values for 'temperature' for site "
                "{}: {}".format(self.site, val)
            )
            logger.error(msg)
            raise InputError(msg)

        val = val.pop()
        logger.debug(
            "Input 'resource_temp' not found in SAM config - setting "
            "to {:.2f} based on input resource data.".format(val)
        )
        self.sam_sys_inputs["resource_temp"] = val

    def _set_egs_plant_design_temperature(self):
        """Set the EGS plant temp to match resource (avoids cf > 1)"""
        if self.sam_sys_inputs.get("resource_type") != 1:
            return  # Not EGS run

        resource_temp = self.sam_sys_inputs["resource_temp"]
        logger.debug("Setting EGS plant design temperature to match "
                     "resource temperature ({}C)".format(resource_temp))
        self.sam_sys_inputs["design_temp"] = resource_temp

    def _set_nameplate_to_match_resource_potential(self, resource):
        """Set the nameplate capacity to match the resource potential."""

        if "nameplate" in self.sam_sys_inputs:
            msg = (
                'Found "nameplate" input in config! Resource potential '
                "from input data will be ignored. Nameplate capacity is "
                "{}".format(self.sam_sys_inputs["nameplate"])
            )
            logger.info(msg)
            # required for downstream LCOE calcs
            self.sam_sys_inputs["system_capacity"] = (
                self.sam_sys_inputs["nameplate"]
            )
            return

        val = set(resource["potential_MW"].unique())
        if len(val) > 1:
            msg = (
                'Found multiple values for "potential_MW" for site '
                "{}: {}".format(self.site, val)
            )
            logger.error(msg)
            raise InputError(msg)

        val = val.pop() * 1000

        logger.debug("Setting the nameplate to {}".format(val))
        self.sam_sys_inputs["nameplate"] = val
        # required for downstream LCOE calcs
        self.sam_sys_inputs["system_capacity"] = val

    def _set_resource_potential_to_match_gross_output(self):
        """Set the resource potential input to match the gross generation.

        If SAM throws an error during the UI calculation of the gross
        output, the resource_potential is simply set to -1 since
        SAM will error out for this point regardless of the
        resource_potential input.
        """

        super().assign_inputs()
        self["ui_calculations_only"] = 1
        try:
            self.execute()
        except SAMExecutionError:
            self["ui_calculations_only"] = 0
            self.sam_sys_inputs["resource_potential"] = -1
            return

        gross_gen = (
            self.pysam.Outputs.gross_output * self._RESOURCE_POTENTIAL_MULT
        )
        if "resource_potential" in self.sam_sys_inputs:
            msg = (
                'Setting "resource_potential" is not allowed! Updating '
                "user input of {} to match the gross generation: {}".format(
                    self.sam_sys_inputs["resource_potential"], gross_gen
                )
            )
            logger.warning(msg)
            warn(msg)

        logger.debug(
            "Setting the resource potential to {} MW".format(gross_gen)
        )
        self.sam_sys_inputs["resource_potential"] = gross_gen

        ncw = self.sam_sys_inputs.pop(
            "num_confirmation_wells", self._DEFAULT_NUM_CONFIRMATION_WELLS
        )
        self.sam_sys_inputs["prod_and_inj_wells_to_drill"] = (
            self.pysam.Outputs.num_wells_getem_output
            - ncw
            + self.pysam.Outputs.num_wells_getem_inj
        )
        self["ui_calculations_only"] = 0

    def _set_costs(self):
        """Set the costs based on gross plant generation."""
        plant_size_kw = (
            self.sam_sys_inputs["resource_potential"]
            / self._RESOURCE_POTENTIAL_MULT
        ) * 1000

        cc_per_kw = self.sam_sys_inputs.pop("capital_cost_per_kw", None)
        if cc_per_kw is not None:
            capital_cost = cc_per_kw * plant_size_kw
            logger.debug(
                "Setting the capital_cost to ${:,.2f}".format(capital_cost)
            )
            reg_mult = self.sam_sys_inputs.get("capital_cost_multiplier", 1)
            self.sam_sys_inputs["base_capital_cost"] = capital_cost
            self.sam_sys_inputs["capital_cost"] = capital_cost * reg_mult

        dc_per_well = self.sam_sys_inputs.pop("drill_cost_per_well", None)
        num_wells = self.sam_sys_inputs.pop(
            "prod_and_inj_wells_to_drill", None
        )
        if dc_per_well is not None:
            if num_wells is None:
                msg = (
                    "Could not determine number of wells to be drilled. "
                    "No drilling costs added!"
                )
                logger.warning(msg)
                warn(msg)
            else:
                capital_cost = self.sam_sys_inputs["capital_cost"]
                drill_cost = dc_per_well * num_wells
                logger.debug(
                    "Setting the drilling cost to ${:,.2f} "
                    "({:.2f} wells at ${:,.2f} per well)".format(
                        drill_cost, num_wells, dc_per_well
                    )
                )
                reg_mult = self.sam_sys_inputs.get(
                    "capital_cost_multiplier", 1
                )
                base_cc = capital_cost / reg_mult
                new_base_cc = base_cc + drill_cost
                self.sam_sys_inputs["base_capital_cost"] = new_base_cc
                self.sam_sys_inputs["capital_cost"] = new_base_cc * reg_mult

        foc_per_kw = self.sam_sys_inputs.pop(
            "fixed_operating_cost_per_kw", None
        )
        if foc_per_kw is not None:
            foc = foc_per_kw * plant_size_kw
            logger.debug(
                "Setting the fixed_operating_cost to ${:,.2f}".format(foc)
            )
            self.sam_sys_inputs["base_fixed_operating_cost"] = foc
            self.sam_sys_inputs["fixed_operating_cost"] = foc

        voc_per_kw = self.sam_sys_inputs.pop(
            "variable_operating_cost_per_kw", None
        )
        if voc_per_kw is not None:
            voc = voc_per_kw * plant_size_kw
            logger.debug(
                "Setting the variable_operating_cost to ${:,.2f}".format(voc)
            )
            self.sam_sys_inputs["base_variable_operating_cost"] = voc
            self.sam_sys_inputs["variable_operating_cost"] = voc

    def _create_pysam_wfile(self, resource, meta):
        """Create PySAM weather input file.

        Geothermal module requires a weather file, but does not actually
        require any weather data to run. Therefore, an empty file is
        generated and passed through.

        Parameters
        ----------
        resource : pd.DataFrame
            Time series resource data for a single location with a
            pandas DatetimeIndex. There must be columns for all the
            required variables to run the respective SAM simulation.
        meta : pd.Series
            Meta data corresponding to the resource input for the single
            location. Should include values for latitude, longitude,
            and timezone.

        Returns
        -------
        fname : str
            Name of weather csv file.

        Notes
        -----
        PySAM will not accept data on Feb 29th. For leap years,
        December 31st is dropped and time steps are shifted to relabel
        Feb 29th as March 1st, March 1st as March 2nd, etc.
        """
        # pylint: disable=attribute-defined-outside-init, consider-using-with
        self._temp_dir = TemporaryDirectory()
        fname = os.path.join(self._temp_dir.name, "weather.csv")
        logger.debug("Creating PySAM weather data file: {}".format(fname))

        # ------- Process metadata
        m = pd.DataFrame(meta).T
        m = m.rename(
            {
                "latitude": "Latitude",
                "longitude": "Longitude",
                "timezone": "Time Zone",
            },
            axis=1,
        )

        m[["Latitude", "Longitude", "Time Zone"]].to_csv(
            fname, index=False, mode="w"
        )

        # --------- Process data, blank for geothermal
        time_index = resource.index
        mask = (time_index.month == 2) & (time_index.day == 29)
        time_index = time_index[~mask]

        df = pd.DataFrame(index=time_index)
        df["Year"] = time_index.year
        df["Month"] = time_index.month
        df["Day"] = time_index.day
        df["Hour"] = time_index.hour
        df["Minute"] = time_index.minute
        df.to_csv(fname, index=False, mode="a")

        return fname

    def run_gen_and_econ(self):
        """Run SAM generation and possibility follow-on econ analysis."""
        try:
            super().run_gen_and_econ()
        except SAMExecutionError as e:
            logger.error(
                "Skipping site {}; received sam error: {}".format(
                    self._site, str(e)
                )
            )
            self.outputs = {}


class AbstractSamWind(AbstractSamGeneration, PowerCurveLossesMixin, ABC):
    """AbstractSamWind"""

    # pylint: disable=line-too-long
    def __init__(self, *args, **kwargs):
        """Wind generation from SAM.

        See the PySAM :py:class:`~PySAM.Windpower.Windpower`
        documentation for the configuration keys required in the
        `sam_sys_inputs` config. You may also include the following
        ``reV``-specific keys:

            - ``reV_power_curve_losses`` : A dictionary that can be used
              to initialize
              :class:`~reV.losses.power_curve.PowerCurveLossesInput`.
              For example::

                    reV_power_curve_losses = {
                        'target_losses_percent': 9.8,
                        'transformation': 'exponential_stretching'
                    }

              See the description of the class mentioned above or the
              `reV losses demo notebook <https://tinyurl.com/4d7uutt3/>`_
              for detailed instructions on how to specify this input.
            - ``reV_outages`` : Specification for ``reV``-scheduled
              stochastic outage losses. For example::

                    outage_info = [
                        {
                            'count': 6,
                            'duration': 24,
                            'percentage_of_capacity_lost': 100,
                            'allowed_months': ['January', 'March'],
                            'allow_outage_overlap': True
                        },
                        {
                            'count': 10,
                            'duration': 1,
                            'percentage_of_capacity_lost': 10,
                            'allowed_months': ['January'],
                            'allow_outage_overlap': False
                        },
                        ...
                    ]

              See the description of
              :meth:`~reV.losses.scheduled.ScheduledLossesMixin.add_scheduled_losses`
              or the
              `reV losses demo notebook <https://tinyurl.com/4d7uutt3/>`_
              for detailed instructions on how to specify this input.
            - ``reV_outages_seed`` : Integer value used to seed the RNG
              used to compute stochastic outage losses.
            - ``time_index_step`` : Integer representing the step size
              used to sample the ``time_index`` in the resource data.
              This can be used to reduce temporal resolution (i.e. for
              30 minute input data, ``time_index_step=1`` yields the
              full 30 minute time series as output, while
              ``time_index_step=2`` yields hourly output, and so forth).

              .. Note:: The reduced data shape (i.e. after applying a
                        step size of `time_index_step`) must still be
                        an integer multiple of 8760, or the execution
                        will fail.

        Parameters
        ----------
        resource : pd.DataFrame
            Timeseries solar or wind resource data for a single location with a
            pandas DatetimeIndex.  There must be columns for all the required
            variables to run the respective SAM simulation. Remapping will be
            done to convert typical NSRDB/WTK names into SAM names (e.g. DNI ->
            dn and wind_speed -> windspeed)
        meta : pd.DataFrame | pd.Series
            Meta data corresponding to the resource input for the single
            location. Should include values for latitude, longitude, elevation,
            and timezone.
        sam_sys_inputs : dict
            Site-agnostic SAM system model inputs arguments.
        site_sys_inputs : dict
            Optional set of site-specific SAM system inputs to complement the
            site-agnostic inputs.
        output_request : list
            Requested SAM outputs (e.g., 'cf_mean', 'annual_energy',
            'cf_profile', 'gen_profile', 'energy_yield', 'ppa_price',
            'lcoe_fcr').
        drop_leap : bool
            Drops February 29th from the resource data. If False, December
            31st is dropped from leap years.
        """
        super().__init__(*args, **kwargs)
        self.add_power_curve_losses()


class WindPower(AbstractSamWind):
    """Class for Wind generation from SAM"""

    MODULE = "windpower"
    PYSAM = PySamWindPower

    def set_resource_data(self, resource, meta):
        """Set WTK resource data arrays.

        Parameters
        ----------
        resource : pd.DataFrame
            Timeseries solar or wind resource data for a single location with a
            pandas DatetimeIndex.  There must be columns for all the required
            variables to run the respective SAM simulation. Remapping will be
            done to convert typical NSRDB/WTK names into SAM names (e.g. DNI ->
            dn and wind_speed -> windspeed)
        meta : pd.Series
            Meta data corresponding to the resource input for the single
            location. Should include values for latitude, longitude, elevation,
            and timezone.
        """

        meta = self._parse_meta(meta)

        # map resource data names to SAM required data names
        var_map = {
            "speed": "windspeed",
            "direction": "winddirection",
            "airtemperature": "temperature",
            "temp": "temperature",
            "surfacepressure": "pressure",
            "relativehumidity": "rh",
            "humidity": "rh",
        }
        lower_case = {
            k: k.lower().replace(" ", "").replace("_", "")
            for k in resource.columns
        }
        resource = resource.rename(mapper=lower_case, axis="columns")
        resource = resource.rename(mapper=var_map, axis="columns")

        data_dict = {}
        var_list = ["temperature", "pressure", "windspeed", "winddirection"]
        if "winddirection" not in resource:
            resource["winddirection"] = 0.0

        time_index = resource.index
        self.time_interval = self.get_time_interval(resource.index.values)

        data_dict["fields"] = [1, 2, 3, 4]
        data_dict["heights"] = 4 * [self.sam_sys_inputs["wind_turbine_hub_ht"]]

        if "rh" in resource:
            # set relative humidity for icing.
            rh = self.ensure_res_len(resource["rh"].values, time_index)
            n_roll = int(meta[ResourceMetaField.TIMEZONE] * self.time_interval)
            rh = np.roll(rh, n_roll, axis=0)
            data_dict["rh"] = rh.tolist()

        # must be set as matrix in [temperature, pres, speed, direction] order
        # ensure that resource array length is multiple of 8760
        # roll the truncated resource array to local timezone
        temp = self.ensure_res_len(resource[var_list].values, time_index)
        n_roll = int(meta[ResourceMetaField.TIMEZONE] * self.time_interval)
        temp = np.roll(temp, n_roll, axis=0)
        data_dict["data"] = temp.tolist()

        data_dict["lat"] = float(meta[ResourceMetaField.LATITUDE])
        data_dict["lon"] = float(meta[ResourceMetaField.LONGITUDE])
        data_dict["tz"] = int(meta[ResourceMetaField.TIMEZONE])
        data_dict["elev"] = float(meta[ResourceMetaField.ELEVATION])

        time_index = self.ensure_res_len(time_index, time_index)
        data_dict["minute"] = time_index.minute.tolist()
        data_dict["hour"] = time_index.hour.tolist()
        data_dict["year"] = time_index.year.tolist()
        data_dict["month"] = time_index.month.tolist()
        data_dict["day"] = time_index.day.tolist()

        # add resource data to self.data and clear
        self["wind_resource_data"] = data_dict
        self["wind_resource_model_choice"] = 0

    @staticmethod
    def default():
        """Get the executed default pysam WindPower object.

        Returns
        -------
        PySAM.Windpower
        """
        return DefaultWindPower.default()


# pylint: disable=too-many-ancestors
class WindPowerPD(AbstractSamGeneration, PowerCurveLossesMixin):
    """WindPower analysis with wind speed/direction joint probabilty
    distrubtion input"""

    MODULE = "windpower"
    PYSAM = PySamWindPower

    def __init__(
        self,
        ws_edges,
        wd_edges,
        wind_dist,
        meta,
        sam_sys_inputs,
        site_sys_inputs=None,
        output_request=None,
    ):
        """Initialize a SAM generation object for windpower with a
        speed/direction joint probability distribution.

        Parameters
        ----------
        ws_edges : np.ndarray
            1D array of windspeed (m/s) values that set the bin edges for the
            wind probability distribution. Same len as wind_dist.shape[0] + 1
        wd_edges : np.ndarray
            1D array of winddirections (deg) values that set the bin edges
            for the wind probability dist. Same len as wind_dist.shape[1] + 1
        wind_dist : np.ndarray
            2D array probability distribution of (windspeed, winddirection).
        meta : pd.DataFrame | pd.Series
            Meta data corresponding to the resource input for the single
            location. Should include values for latitude, longitude, elevation,
            and timezone.
        sam_sys_inputs : dict
            Site-agnostic SAM system model inputs arguments.
        site_sys_inputs : dict
            Optional set of site-specific SAM system inputs to complement the
            site-agnostic inputs.
        output_request : list
            Requested SAM outputs (e.g., 'cf_mean', 'annual_energy',
            'cf_profile', 'gen_profile', 'energy_yield', 'ppa_price',
            'lcoe_fcr').
        """

        # make sure timezone and elevation are in the meta data
        meta = self.tz_elev_check(sam_sys_inputs, site_sys_inputs, meta)

        # don't pass resource to base class,
        # set in concrete generation classes instead
        super().__init__(
            None,
            meta,
            sam_sys_inputs,
            site_sys_inputs=site_sys_inputs,
            output_request=output_request,
            drop_leap=False,
        )

        # Set the site number using meta data
        if hasattr(meta, "name"):
            self._site = meta.name
        else:
            self._site = None

        self.set_resource_data(ws_edges, wd_edges, wind_dist)
        self.add_power_curve_losses()

    def set_resource_data(self, ws_edges, wd_edges, wind_dist):
        """Send wind PD to pysam

        Parameters
        ----------
        ws_edges : np.ndarray
            1D array of windspeed (m/s) values that set the bin edges for the
            wind probability distribution. Same len as wind_dist.shape[0] + 1
        wd_edges : np.ndarray
            1D array of winddirections (deg) values that set the bin edges
            for the wind probability dist. Same len as wind_dist.shape[1] + 1
        wind_dist : np.ndarray
            2D array probability distribution of (windspeed, winddirection).
        """

        assert len(ws_edges) == wind_dist.shape[0] + 1
        assert len(wd_edges) == wind_dist.shape[1] + 1

        wind_dist /= wind_dist.sum()

        # SAM wants the midpoints of the sample bins
        ws_points = ws_edges[:-1] + np.diff(ws_edges) / 2
        wd_points = wd_edges[:-1] + np.diff(wd_edges) / 2

        wd_points, ws_points = np.meshgrid(wd_points, ws_points)
        vstack = (
            ws_points.flatten(),
            wd_points.flatten(),
            wind_dist.flatten(),
        )
        wrd = np.vstack(vstack).T.tolist()

        self["wind_resource_model_choice"] = 2
        self["wind_resource_distribution"] = wrd


class MhkWave(AbstractSamGeneration):
    """Class for Wave generation from SAM"""

    MODULE = "mhkwave"
    PYSAM = PySamMhkWave

    def set_resource_data(self, resource, meta):
        """Set Hindcast US Wave resource data arrays.

        Parameters
        ----------
        resource : pd.DataFrame
            Timeseries resource data for a single location with a
            pandas DatetimeIndex. There must be columns for all the required
            variables to run the respective SAM simulation.
        meta : pd.Series
            Meta data corresponding to the resource input for the single
            location. Should include values for latitude, longitude, elevation,
            and timezone.
        """

        meta = self._parse_meta(meta)

        # map resource data names to SAM required data names
        var_map = {
            "significantwaveheight": "significant_wave_height",
            "waveheight": "significant_wave_height",
            "height": "significant_wave_height",
            "swh": "significant_wave_height",
            "energyperiod": "energy_period",
            "waveperiod": "energy_period",
            "period": "energy_period",
            "ep": "energy_period",
        }
        lower_case = {
            k: k.lower().replace(" ", "").replace("_", "")
            for k in resource.columns
        }
        resource = resource.rename(mapper=lower_case, axis="columns")
        resource = resource.rename(mapper=var_map, axis="columns")

        data_dict = {}

        time_index = resource.index
        self.time_interval = self.get_time_interval(resource.index.values)

        # must be set as matrix in [temperature, pres, speed, direction] order
        # ensure that resource array length is multiple of 8760
        # roll the truncated resource array to local timezone
        for var in ["significant_wave_height", "energy_period"]:
            arr = self.ensure_res_len(resource[var].values, time_index)
            n_roll = int(meta[ResourceMetaField.TIMEZONE] * self.time_interval)
            data_dict[var] = np.roll(arr, n_roll, axis=0).tolist()

        data_dict["lat"] = meta[ResourceMetaField.LATITUDE]
        data_dict["lon"] = meta[ResourceMetaField.LONGITUDE]
        data_dict["tz"] = meta[ResourceMetaField.TIMEZONE]

        time_index = self.ensure_res_len(time_index, time_index)
        data_dict["minute"] = time_index.minute
        data_dict["hour"] = time_index.hour
        data_dict["year"] = time_index.year
        data_dict["month"] = time_index.month
        data_dict["day"] = time_index.day

        # add resource data to self.data and clear
        self["wave_resource_data"] = data_dict

    @staticmethod
    def default():
        """Get the executed default PySAM MhkWave object.

        Returns
        -------
        PySAM.MhkWave
        """
        return DefaultMhkWave.default()



# -*- coding: utf-8 -*-
"""SAM Wind Balance of System Cost Model"""

from copy import deepcopy

import numpy as np
from PySAM.PySSC import ssc_sim_from_dict

from reV.utilities.exceptions import SAMInputError


class WindBos:
    """Wind Balance of System Cost Model."""

    MODULE = "windbos"

    # keys for the windbos input data dictionary.
    # Some keys may not be found explicitly in the SAM input.
    KEYS = (
        "tech_model",
        "financial_model",
        "machine_rating",
        "rotor_diameter",
        "hub_height",
        "number_of_turbines",
        "interconnect_voltage",
        "distance_to_interconnect",
        "site_terrain",
        "turbine_layout",
        "soil_condition",
        "construction_time",
        "om_building_size",
        "quantity_test_met_towers",
        "quantity_permanent_met_towers",
        "weather_delay_days",
        "crane_breakdowns",
        "access_road_entrances",
        "turbine_capital_cost",
        "turbine_cost_per_kw",
        "tower_top_mass",
        "delivery_assist_required",
        "pad_mount_transformer_required",
        "new_switchyard_required",
        "rock_trenching_required",
        "mv_thermal_backfill",
        "mv_overhead_collector",
        "performance_bond",
        "contingency",
        "warranty_management",
        "sales_and_use_tax",
        "overhead",
        "profit_margin",
        "development_fee",
        "turbine_transportation",
    )

    def __init__(self, inputs):
        """
        Parameters
        ----------
        inputs : dict
            SAM key value pair inputs.
        """

        self._turbine_capital_cost = 0.0
        self._datadict = {}

        self._inputs = inputs
        self._special = {
            "tech_model": "windbos",
            "financial_model": "none",
            "machine_rating": self.machine_rating,
            "hub_height": self.hub_height,
            "rotor_diameter": self.rotor_diameter,
            "number_of_turbines": self.number_of_turbines,
            "turbine_capital_cost": self.turbine_capital_cost,
        }
        self._parse_inputs()
        self._out = ssc_sim_from_dict(self._datadict)

    def _parse_inputs(self):
        """Parse SAM inputs into a windbos input dict and perform any
        required special operations."""

        for k in self.KEYS:
            if k in self._special:
                self._datadict[k] = self._special[k]
            elif k not in self._inputs:
                raise SAMInputError(
                    'Windbos requires input key: "{}"'.format(k)
                )
            else:
                self._datadict[k] = self._inputs[k]

    @property
    def machine_rating(self):
        """Single turbine machine rating either from input or power curve."""
        if "machine_rating" in self._inputs:
            return self._inputs["machine_rating"]
        else:
            return np.max(self._inputs["wind_turbine_powercurve_powerout"])

    @property
    def hub_height(self):
        """Turbine hub height."""
        if "wind_turbine_hub_ht" in self._inputs:
            return self._inputs["wind_turbine_hub_ht"]
        else:
            return self._inputs["hub_height"]

    @property
    def rotor_diameter(self):
        """Turbine rotor diameter."""
        if "wind_turbine_rotor_diameter" in self._inputs:
            return self._inputs["wind_turbine_rotor_diameter"]
        else:
            return self._inputs["rotor_diameter"]

    @property
    def number_of_turbines(self):
        """Number of turbines either based on input or system (farm) capacity
        and machine rating"""

        if "number_of_turbines" in self._inputs:
            return self._inputs["number_of_turbines"]
        else:
            return (
                self._inputs['system_capacity'] / self.machine_rating
            )

    @property
    def turbine_capital_cost(self):
        """Returns zero (no turbine capital cost for WindBOS input,
        and assigns any input turbine_capital_cost to an attr"""

        if "turbine_capital_cost" in self._inputs:
            self._turbine_capital_cost = self._inputs["turbine_capital_cost"]
        else:
            self._turbine_capital_cost = 0.0
        return 0.0

    @property
    def bos_cost(self):
        """Get the balance of system cost ($)."""
        return self._out["project_total_budgeted_cost"]

    @property
    def turbine_cost(self):
        """Get the turbine cost ($)."""
        tcost = (
            self._inputs["turbine_cost_per_kw"]
            * self.machine_rating
            * self.number_of_turbines
        ) + (self._turbine_capital_cost * self.number_of_turbines)
        return tcost

    @property
    def sales_tax_mult(self):
        """Get a sales tax multiplier (frac of the total installed cost)."""
        basis = self._inputs.get("sales_tax_basis", 0) / 100
        tax = self._datadict.get("sales_and_use_tax", 0) / 100
        return basis * tax

    @property
    def sales_tax_cost(self):
        """Get the cost of sales tax ($)."""
        return (self.bos_cost + self.turbine_cost) * self.sales_tax_mult

    @property
    def total_installed_cost(self):
        """Get the total installed cost ($) (bos + turbine)."""
        return self.bos_cost + self.turbine_cost + self.sales_tax_cost

    @property
    def output(self):
        """Get a dictionary containing the cost breakdown."""
        output = {
            "total_installed_cost": self.total_installed_cost,
            "turbine_cost": self.turbine_cost,
            "sales_tax_cost": self.sales_tax_cost,
            "bos_cost": self.bos_cost,
        }
        return output

    # pylint: disable-msg=W0613
    @classmethod
    def reV_run(
        cls,
        points_control,
        site_df,
        output_request=("total_installed_cost",),
        **kwargs,
    ):
        """Execute SAM SingleOwner simulations based on reV points control.

        Parameters
        ----------
        points_control : config.PointsControl
            PointsControl instance containing project points site and SAM
            config info.
        site_df : pd.DataFrame
            Dataframe of site-specific input variables. Row index corresponds
            to site number/gid (via df.loc not df.iloc), column labels are the
            variable keys that will be passed forward as SAM parameters.
        output_request : list | tuple | str
            Output(s) to retrieve from SAM.
        kwargs : dict
            Not used but maintained for polymorphic calls with other
            SAM econ reV_run() methods (lcoe and single owner).
            Breaks pylint error W0613: unused argument.

        Returns
        -------
        out : dict
            Nested dictionaries where the top level key is the site index,
            the second level key is the variable name, second level value is
            the output variable value.
        """
        out = {}

        for site in points_control.sites:
            # get SAM inputs from project_points based on the current site
            _, inputs = points_control.project_points[site]

            # ensure that site-specific data is not persisted to other sites
            site_inputs = deepcopy(inputs)

            site_inputs.update(dict(site_df.loc[site, :]))

            wb = cls(site_inputs)

            out[site] = {
                k: v for k, v in wb.output.items() if k in output_request
            }

        return out



# -*- coding: utf-8 -*-
"""
reV quality assurance and control classes
"""

import logging
import os
from warnings import warn

import numpy as np
import pandas as pd
from gaps.status import Status

from reV.qa_qc.summary import (
    ExclusionsMask,
    SummarizeH5,
    SummarizeSupplyCurve,
    SummaryPlots,
    SupplyCurvePlot,
)
from reV.supply_curve.exclusions import ExclusionMaskFromDict
from reV.utilities import ModuleName, SupplyCurveField, log_versions
from reV.utilities.exceptions import PipelineError

logger = logging.getLogger(__name__)


class QaQc:
    """
    reV QA/QC
    """

    def __init__(self, out_dir):
        """
        Parameters
        ----------
        out_dir : str
            Directory path to save summary data and plots too
        """
        log_versions(logger)
        logger.info("QA/QC results to be saved to: {}".format(out_dir))
        if not os.path.exists(out_dir):
            os.makedirs(out_dir, exist_ok=True)

        self._out_dir = out_dir

    @property
    def out_dir(self):
        """
        Output directory

        Returns
        -------
        str
        """
        return self._out_dir

    @staticmethod
    def _scatter_plot(
        summary_csv, out_root, plot_type="plotly", cmap="viridis", **kwargs
    ):
        """
        Create scatter plot for all summary stats in summary table and save to
        out_dir

        Parameters
        ----------
        summary_csv : str
            Path to .csv file containing summary table
        out_root : str
            Output directory to save plots to
        plot_type : str, optional
            plot_type of plot to create 'plot' or 'plotly', by default 'plotly'
        cmap : str, optional
            Colormap name, by default 'viridis'
        kwargs : dict
            Additional plotting kwargs
        """
        out_dir = os.path.join(
            out_root, os.path.basename(summary_csv).rstrip(".csv")
        )
        if not os.path.exists(out_dir):
            os.makedirs(out_dir, exist_ok=True)

        SummaryPlots.scatter_all(
            summary_csv, out_dir, plot_type=plot_type, cmap=cmap, **kwargs
        )

    def create_scatter_plots(
        self, plot_type="plotly", cmap="viridis", **kwargs
    ):
        """
        Create scatter plot for all compatible summary .csv files

        Parameters
        ----------
        plot_type : str, optional
            plot_type of plot to create 'plot' or 'plotly', by default 'plotly'
        cmap : str, optional
            Colormap name, by default 'viridis'
        kwargs : dict
            Additional plotting kwargs
        """
        for file in os.listdir(self.out_dir):
            if file.endswith(".csv"):
                summary_csv = os.path.join(self.out_dir, file)
                summary = pd.read_csv(summary_csv)
                has_right_cols = ("gid" in summary
                                  and SupplyCurveField.LATITUDE in summary
                                  and SupplyCurveField.LONGITUDE in summary)
                if has_right_cols:
                    self._scatter_plot(summary_csv, self.out_dir,
                                       plot_type=plot_type, cmap=cmap,
                                       **kwargs)

    @classmethod
    def h5(
        cls,
        h5_file,
        out_dir,
        dsets=None,
        group=None,
        process_size=None,
        max_workers=None,
        plot_type="plotly",
        cmap="viridis",
        **kwargs,
    ):
        """
        Run QA/QC by computing summary stats from dsets in h5_file and
        plotting scatters plots of compatible summary stats

        Parameters
        ----------
        h5_file : str
            Path to .h5 file to run QA/QC on
        out_dir : str
            Directory path to save summary tables and plots too
        dsets : str | list, optional
            Datasets to summarize, by default None
        group : str, optional
            Group within h5_file to summarize datasets for, by default None
        process_size : int, optional
            Number of sites to process at a time, by default None
        max_workers : int, optional
            Number of workers to use when summarizing 2D datasets,
            by default None
        plot_type : str, optional
            plot_type of plot to create 'plot' or 'plotly', by default 'plotly'
        cmap : str, optional
            Colormap name, by default 'viridis'
        kwargs : dict
            Additional plotting kwargs
        """
        try:
            qa_qc = cls(out_dir)
            SummarizeH5.run(
                h5_file,
                out_dir,
                group=group,
                dsets=dsets,
                process_size=process_size,
                max_workers=max_workers,
            )
            qa_qc.create_scatter_plots(
                plot_type=plot_type, cmap=cmap, **kwargs
            )
        except Exception as e:
            logger.exception(
                "QAQC failed on file: {}. Received exception:\n{}".format(
                    os.path.basename(h5_file), e
                )
            )
            raise e
        else:
            logger.info(
                "Finished QAQC on file: {} output directory: {}".format(
                    os.path.basename(h5_file), out_dir
                )
            )

    @classmethod
    def supply_curve(cls, sc_table, out_dir, columns=None,
                     lcoe=SupplyCurveField.MEAN_LCOE, plot_type='plotly',
                     cmap='viridis', sc_plot_kwargs=None,
                     scatter_plot_kwargs=None):
        """
        Plot supply curve

        Parameters
        ----------
        sc_table : str
            Path to .csv file containing supply curve table
        out_dir : str
            Directory path to save summary tables and plots too
        columns : str | list, optional
            Column(s) to summarize, if None summarize all numeric columns,
            by default None
        lcoe : str, optional
            LCOE value to plot, by default :obj:`SupplyCurveField.MEAN_LCOE`
        plot_type : str, optional
            plot_type of plot to create 'plot' or 'plotly', by default 'plotly'
        cmap : str, optional
            Colormap name, by default 'viridis'
        sc_plot_kwargs : dict, optional
            Kwargs for supply curve plot, by default None
        scatter_plot_kwargs : dict
            Kwargs for scatter plot, by default None
        """
        if sc_plot_kwargs is None:
            sc_plot_kwargs = {}

        if scatter_plot_kwargs is None:
            scatter_plot_kwargs = {}

        try:
            qa_qc = cls(out_dir)
            SummarizeSupplyCurve.run(sc_table, out_dir, columns=columns)
            SupplyCurvePlot.plot(
                sc_table,
                out_dir,
                plot_type=plot_type,
                lcoe=lcoe,
                **sc_plot_kwargs,
            )
            qa_qc._scatter_plot(
                sc_table,
                out_dir,
                plot_type=plot_type,
                cmap=cmap,
                **scatter_plot_kwargs,
            )
        except Exception as e:
            logger.exception(
                "QAQC failed on file: {}. Received exception:\n{}".format(
                    os.path.basename(sc_table), e
                )
            )
            raise e
        else:
            logger.info(
                "Finished QAQC on file: {} output directory: {}".format(
                    os.path.basename(sc_table), out_dir
                )
            )

    @classmethod
    def exclusions_mask(
        cls,
        excl_h5,
        out_dir,
        layers_dict=None,
        min_area=None,
        kernel="queen",
        hsds=False,
        plot_type="plotly",
        cmap="viridis",
        plot_step=100,
        **kwargs,
    ):
        """
        Create inclusion mask from given layers dictionary, dump to disk and
        plot

        Parameters
        ----------
        excl_h5 : str
            Path to exclusions .h5 file
        layers_dict : dict | NoneType
            Dictionary of LayerMask arugments {layer: {kwarg: value}}
        min_area : float | NoneType
            Minimum required contiguous area in sq-km
        kernel : str
            Contiguous filter method to use on final exclusions
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS
        plot_type : str, optional
            plot_type of plot to create 'plot' or 'plotly', by default 'plotly'
        cmap : str, optional
            Colormap name, by default 'viridis'
        plot_step : int
            Step between points to plot
        kwargs : dict
            Additional plotting kwargs
        """
        try:
            cls(out_dir)
            excl_mask = ExclusionMaskFromDict.run(
                excl_h5,
                layers_dict=layers_dict,
                min_area=min_area,
                kernel=kernel,
                hsds=hsds,
            )
            excl_mask = np.round(excl_mask * 100).astype("uint8")

            out_file = os.path.basename(excl_h5).replace(".h5", "_mask.npy")
            out_file = os.path.join(out_dir, out_file)
            np.save(out_file, excl_mask)

            ExclusionsMask.plot(
                excl_mask,
                out_dir,
                plot_type=plot_type,
                cmap=cmap,
                plot_step=plot_step,
                **kwargs,
            )
        except Exception as e:
            logger.exception(
                "QAQC failed on file: {}. Received exception:\n{}".format(
                    os.path.basename(excl_h5), e
                )
            )
            raise e
        else:
            logger.info(
                "Finished QAQC on file: {} output directory: {}".format(
                    os.path.basename(excl_h5), out_dir
                )
            )


class QaQcModule:
    """Class to handle Module QA/QC"""

    def __init__(self, module_name, config, out_root):
        """
        Parameters
        ----------
        config : dict
            Dictionary with pre-extracted config input group.
        """
        if not isinstance(config, dict):
            raise TypeError(
                "Config input must be a dict but received: {}".format(
                    type(config)
                )
            )

        self._name = module_name
        self._config = config
        self._out_root = out_root
        self._default_plot_type = "plotly"
        self._default_cmap = "viridis"
        self._default_plot_step = 100
        self._default_lcoe = SupplyCurveField.MEAN_LCOE
        self._default_area_filter_kernel = 'queen'

    @property
    def fpath(self):
        """Get the reV module output filepath(s)

        Returns
        -------
        fpaths : str | list
            One or more filepaths output by current module being QA'd
        """

        fpath = self._config["fpath"]

        if fpath == "PIPELINE":
            target_modules = [self._name]
            for target_module in target_modules:
                fpath = Status.parse_step_status(self._out_root, target_module)
                if fpath:
                    break
            else:
                raise PipelineError(
                    "Could not parse fpath from previous pipeline jobs."
                )
            fpath = fpath[0]
            logger.info(
                "QA/QC using the following "
                "pipeline input for fpath: {}".format(fpath)
            )

        return fpath

    @property
    def sub_dir(self):
        """
        QA/QC sub directory for this module's outputs
        """
        return self._config.get("sub_dir", None)

    @property
    def plot_type(self):
        """Get the QA/QC plot type: either 'plot' or 'plotly'"""
        return self._config.get("plot_type", self._default_plot_type)

    @property
    def dsets(self):
        """Get the reV_h5 dsets to QA/QC"""
        return self._config.get("dsets", None)

    @property
    def group(self):
        """Get the reV_h5 group to QA/QC"""
        return self._config.get("group", None)

    @property
    def process_size(self):
        """Get the reV_h5 process_size for QA/QC"""
        return self._config.get("process_size", None)

    @property
    def cmap(self):
        """Get the QA/QC plot colormap"""
        return self._config.get("cmap", self._default_cmap)

    @property
    def plot_step(self):
        """Get the QA/QC step between exclusion mask points to plot"""
        return self._config.get("cmap", self._default_plot_step)

    @property
    def columns(self):
        """Get the supply_curve columns to QA/QC"""
        return self._config.get("columns", None)

    @property
    def lcoe(self):
        """Get the supply_curve lcoe column to plot"""
        return self._config.get("lcoe", self._default_lcoe)

    @property
    def excl_fpath(self):
        """Get the source exclusions filepath"""
        excl_fpath = self._config.get("excl_fpath", "PIPELINE")

        if excl_fpath == "PIPELINE":
            target_module = ModuleName.SUPPLY_CURVE_AGGREGATION
            excl_fpath = Status.parse_step_status(
                self._out_root, target_module, key="excl_fpath"
            )
            if not excl_fpath:
                excl_fpath = None
                msg = (
                    "Could not parse excl_fpath from previous "
                    "pipeline jobs, defaulting to: {}".format(excl_fpath)
                )
                logger.warning(msg)
                warn(msg)
            else:
                excl_fpath = excl_fpath[0]
                logger.info(
                    "QA/QC using the following "
                    "pipeline input for excl_fpath: {}".format(excl_fpath)
                )

        return excl_fpath

    @property
    def excl_dict(self):
        """Get the exclusions dictionary"""
        excl_dict = self._config.get("excl_dict", "PIPELINE")

        if excl_dict == "PIPELINE":
            target_module = ModuleName.SUPPLY_CURVE_AGGREGATION
            excl_dict = Status.parse_step_status(
                self._out_root, target_module, key="excl_dict"
            )
            if not excl_dict:
                excl_dict = None
                msg = (
                    "Could not parse excl_dict from previous "
                    "pipeline jobs, defaulting to: {}".format(excl_dict)
                )
                logger.warning(msg)
                warn(msg)
            else:
                excl_dict = excl_dict[0]
                logger.info(
                    "QA/QC using the following "
                    "pipeline input for excl_dict: {}".format(excl_dict)
                )

        return excl_dict

    @property
    def area_filter_kernel(self):
        """Get the minimum area filter kernel name ('queen' or 'rook')."""
        area_filter_kernel = self._config.get("area_filter_kernel", "PIPELINE")

        if area_filter_kernel == "PIPELINE":
            target_module = ModuleName.SUPPLY_CURVE_AGGREGATION
            key = "area_filter_kernel"
            area_filter_kernel = Status.parse_step_status(
                self._out_root, target_module, key=key
            )
            if not area_filter_kernel:
                area_filter_kernel = self._default_area_filter_kernel
                msg = (
                    "Could not parse area_filter_kernel from previous "
                    "pipeline jobs, defaulting to: {}".format(
                        area_filter_kernel
                    )
                )
                logger.warning(msg)
                warn(msg)
            else:
                area_filter_kernel = area_filter_kernel[0]
                logger.info(
                    "QA/QC using the following "
                    "pipeline input for area_filter_kernel: {}".format(
                        area_filter_kernel
                    )
                )

        return area_filter_kernel

    @property
    def min_area(self):
        """Get the minimum area filter minimum area in km2."""
        min_area = self._config.get("min_area", "PIPELINE")

        if min_area == "PIPELINE":
            target_module = ModuleName.SUPPLY_CURVE_AGGREGATION
            min_area = Status.parse_step_status(
                self._out_root, target_module, key="min_area"
            )
            if not min_area:
                min_area = None
                msg = (
                    "Could not parse min_area from previous "
                    "pipeline jobs, defaulting to: {}".format(min_area)
                )
                logger.warning(msg)
                warn(msg)
            else:
                min_area = min_area[0]
                logger.info(
                    "QA/QC using the following "
                    "pipeline input for min_area: {}".format(min_area)
                )

        return min_area



# -*- coding: utf-8 -*-
"""
Generate reV inclusion mask from exclusion layers
"""
import logging
from warnings import warn

import numpy as np
from rex.utilities.loggers import log_mem
from scipy import ndimage

from reV.handlers.exclusions import ExclusionLayers, LATITUDE, LONGITUDE
from reV.utilities.exceptions import ExclusionLayerError, SupplyCurveInputError

logger = logging.getLogger(__name__)


class LayerMask:
    """
    Class to convert exclusion layer to inclusion layer mask
    """

    def __init__(self, layer,
                 exclude_values=None,
                 exclude_range=(None, None),
                 include_values=None,
                 include_range=(None, None),
                 include_weights=None,
                 force_include_values=None,
                 force_include_range=None,
                 use_as_weights=False,
                 weight=1.0,
                 exclude_nodata=False,
                 nodata_value=None,
                 extent=None,
                 **kwargs):
        """
        Parameters
        ----------
        layer : str
            Layer name.
        exclude_values : int | float | list, optional
            Single value or list of values to exclude.

            .. Important:: The keyword arguments `exclude_values`,
              `exclude_range`, `include_values`, `include_range`,
              `include_weights`, `force_include_values`, and
              `force_include_range` are all mutually exclusive. Users
              should supply value(s) for exactly one of these arguments.

            By default, ``None``.
        exclude_range : list | tuple, optional
            Two-item list of [min threshold, max threshold] (ends are
            inclusive) for values to exclude. Mutually exclusive
            with other inputs (see info in the description of
            `exclude_values`). By default, ``None``.
        include_values : int | float | list, optional
            Single value or list of values to include. Mutually
            exclusive with other inputs (see info in the description of
            `exclude_values`). By default, ``None``.
        include_range : list | tuple, optional
            Two-item list of [min threshold, max threshold] (ends are
            inclusive) for values to include. Mutually exclusive with
            other inputs (see info in the description of
            `exclude_values`). By default, ``None``.
        include_weights : dict, optional
            A dictionary of ``{value: weight}`` pairs, where the
            ``value`` in the layer that should be included with the
            given ``weight``. Mutually exclusive with other inputs (see
            info in the description of  `exclude_values`).
            By default, ``None``.
        force_include_values : int | float | list, optional
            Force the inclusion of the given value(s). This input
            completely replaces anything provided as `include_values`
            and is mutually exclusive with other inputs (eee info in
            the description of `exclude_values`). By default, ``None``.
        force_include_range : list | tuple, optional
            Force the inclusion of given values in the range
            [min threshold, max threshold] (ends are inclusive). This
            input completely replaces anything provided as
            `include_range` and is mutually exclusive with other inputs
            (see info in the description of `exclude_values`).
            By default, ``None``.
        use_as_weights : bool, optional
            Option to use layer as final inclusion weights (i.e.
            1 = fully included, 0.75 = 75% included, 0.5 = 50% included,
            etc.). If ``True``, all inclusion/exclusions specifications
            for the layer are ignored and the raw values (scaled by the
            `weight` input) are used as inclusion weights.
            By default, ``False``.
        weight : float, optional
            Weight applied to exclusion layer after it is calculated.
            Can be used, for example, to turn a binary exclusion layer
            (i.e. data with 0 or 1 values and ``exclude_values=1``
            input) into partial exclusions by setting the weight to
            a fraction (e.g. 0.5 for 50% exclusions). By default, ``1``.
        exclude_nodata : bool, optional
            Flag to exclude nodata values (`nodata_value`). If
            ``nodata_value=None`` the `nodata_value` is inferred by
            :class:`reV.supply_curve.exclusions.ExclusionMask`.
            By default, ``False``.
        nodata_value : int | float, optional
            Nodata value for the layer. If ``None``, the value will be
            inferred when LayerMask is added to
            :class:`reV.supply_curve.exclusions.ExclusionMask`.
            By default, ``None``.
        extent : dict, optional
            Optional dictionary with values that can be used to
            initialize this class (i.e. `layer`, `exclude_values`,
            `include_range`, etc.). This dictionary should contain the
            specifications to create a boolean mask that defines the
            extent to which the original mask should be applied.
            For example, suppose you specify the input the following
            way::

                input_dict = {
                    "viewsheds": {
                        "exclude_values": 1,
                        "extent": {
                            "layer": "federal_parks",
                            "include_range": [1, 5]
                        }
                    }
                }

                for layer_name, kwargs in input_dict.items():
                    layer = LayerMask(layer_name, **kwargs)
                    ...

            This would mean that you are masking out all viewshed layer
            values equal to 1, **but only where the "federal_parks"
            layer is equal to 1, 2, 3, 4, or 5**. Outside of these
            regions (i.e. outside of federal park regions), the viewshed
            exclusion is **NOT** applied. If the extent mask created by
            these options is not boolean, an error is thrown (i.e. do
            not specify `weight` or `use_as_weights`).
            By default ``None``, which applies the original layer mask
            to the full extent.
        **kwargs
            Optional inputs to maintain legacy kwargs of ``inclusion_*``
            instead of ``include_*``.
        """

        self._name = layer
        self._exclude_values = exclude_values
        self._exclude_range = exclude_range
        self._include_values = include_values
        self._include_range = include_range
        self._include_weights = include_weights
        self._force_include = False

        self._parse_legacy_kwargs(kwargs)

        if force_include_values is not None:
            self._include_values = force_include_values
            self._force_include = True
        if force_include_range is not None:
            self._include_range = force_include_range
            self._force_include = True

        self._as_weights = use_as_weights
        self._exclude_nodata = exclude_nodata
        self.nodata_value = nodata_value

        if weight > 1 or weight < 0:
            msg = ('Invalid weight ({}) provided for layer {}:'
                   '\nWeight must fall between 0 and 1!'.format(weight, layer))
            logger.error(msg)
            raise ValueError(msg)

        self._weight = weight
        self._mask_type = self._check_mask_type()
        self.extent = LayerMask(**extent) if extent is not None else None

    def __repr__(self):
        msg = ('{} for "{}" exclusion, of type "{}"'
               .format(self.__class__.__name__, self.name, self.mask_type))

        return msg

    def __getitem__(self, data):
        """Get the multiplicative inclusion mask.

        Returns
        -------
        mask : ndarray
            Masked exclusion data with weights applied such that 1 is included,
            0 is excluded, 0.5 is half included.
        """
        return self._apply_mask(data)

    def _parse_legacy_kwargs(self, kwargs):
        """Parse legacy kwargs that start with inclusion_* instead of include_*

        Parameters
        ----------
        kwargs : dict
            Optional inputs to maintain legacy kwargs of inclusion_* instead of
            include_*
        """

        for k, v in kwargs.items():
            msg = None
            if k == 'inclusion_range':
                self._include_range = v
                msg = 'Please use "include_range" instead of "inclusion_range"'

            elif k == 'inclusion_weights':
                self._include_weights = v
                msg = ('Please use "include_weights" instead of '
                       '"inclusion_weights"')

            elif k == 'inclusion_values':
                self._include_values = v
                msg = ('Please use "include_values" instead of '
                       '"inclusion_values"')

            if msg is not None:
                warn(msg)
                logger.warning(msg)

    @property
    def name(self):
        """
        Layer name to extract from exclusions .h5 file

        Returns
        -------
        _name : str
        """
        return self._name

    @property
    def min_value(self):
        """Minimum value to include/exclude if include_range or exclude_range
        was input.

        Returns
        -------
        float
        """
        if 'excl' in self.mask_type:
            range_var = self._exclude_range
        else:
            range_var = self._include_range

        if all(isinstance(x, (int, float)) for x in range_var):
            return min(range_var)
        return range_var[0]

    @property
    def max_value(self):
        """Maximum value to include/exclude if include_range or exclude_range
        was input.

        Returns
        -------
        float
        """
        if 'excl' in self.mask_type:
            range_var = self._exclude_range
        else:
            range_var = self._include_range

        if all(isinstance(x, (int, float)) for x in range_var):
            return max(range_var)
        return range_var[1]

    @property
    def exclude_values(self):
        """
        Values to exclude

        Returns
        -------
        _exclude_values : list
        """
        return self._exclude_values

    @property
    def include_values(self):
        """
        Values to include

        Returns
        -------
        _include_values : list
        """
        return self._include_values

    @property
    def include_weights(self):
        """
        Mapping of values to include and at what weights

        Returns
        -------
        dict
        """
        return self._include_weights

    @property
    def force_include(self):
        """
        Flag to force include mask

        Returns
        -------
        _force_include : bool
        """
        return self._force_include

    @property
    def mask_type(self):
        """
        Type of exclusion mask for this layer

        Returns
        -------
        str
        """
        return self._mask_type

    def _apply_mask(self, data):
        """
        Apply mask function

        Parameters
        ----------
        data : ndarray
            Exclusions data to create mask from

        Returns
        -------
        data : ndarray
            Masked exclusion data with weights applied such that 1 is included,
            0 is excluded, 0.5 is half included.
        """

        if not self._as_weights:
            if self.mask_type == 'include_range':
                func = self._include_range_mask
            elif self.mask_type == 'exclude_range':
                func = self._exclude_range_mask
            elif self.mask_type == 'exclude':
                func = self._exclusion_mask
            elif self.mask_type == 'include':
                func = self._inclusion_mask
            elif self.mask_type == 'include_weights':
                func = self._weights_mask
            else:
                msg = ('{} is an invalid mask type: expecting '
                       '"include_range", "exclude_range", "exclude", '
                       '"include", or "include_weights"'
                       .format(self.mask_type))
                logger.error(msg)
                raise KeyError(msg)

            data = func(data)

        data = data.astype('float32') * self._weight

        return data

    def _check_mask_type(self):
        """
        Ensure that the initialization arguments are valid and not
        contradictory

        Returns
        -------
        mask : str
            Mask type
        """
        mask = None
        if not self._as_weights:
            masks = {'include_range': any(i is not None
                                          for i in self._include_range),
                     'exclude_range': any(i is not None
                                          for i in self._exclude_range),
                     'exclude': self._exclude_values is not None,
                     'include': self._include_values is not None,
                     'include_weights': self._include_weights is not None}
            for k, v in masks.items():
                if v:
                    if mask is None:
                        mask = k
                    else:
                        msg = ('Only one approach can be used to create the '
                               'inclusion mask, but you supplied {} and {}'
                               .format(mask, k))
                        logger.error(msg)
                        raise ExclusionLayerError(msg)

            if mask is None:
                msg = ('Exactly one approach must be specified to create the '
                       'inclusion mask for layer {!r}! Please specify one of: '
                       '`exclude_values`, `exclude_range`, `include_values`, '
                       '`include_range`, `include_weights`, '
                       '`force_include_values`, or `force_include_range`.'
                       .format(self.name))
                logger.error(msg)
                raise ExclusionLayerError(msg)

        if mask == 'include_weights' and self._weight < 1:
            msg = ("Values are individually weighted when using "
                   "'include_weights', the supplied weight of {} will be "
                   "ignored!".format(self._weight))
            self._weight = 1
            logger.warning(msg)
            warn(msg)

        return mask

    def _exclude_range_mask(self, data):
        """
        Mask exclusion layer based on exclude value range

        Parameters
        ----------
        data : ndarray
            Exclusions data to create mask from

        Returns
        -------
        mask : ndarray
            Boolean mask of which values to include (True is include).
        """
        mask = np.full(data.shape, False)
        if self.min_value is not None:
            mask = data < self.min_value

        if self.max_value is not None:
            mask |= data > self.max_value

        mask[data == self.nodata_value] = True
        if self._exclude_nodata:
            mask = mask & (data != self.nodata_value)

        return mask

    def _include_range_mask(self, data):
        """
        Mask exclusion layer based on include value range

        Parameters
        ----------
        data : ndarray
            Exclusions data to create mask from

        Returns
        -------
        mask : ndarray
            Boolean mask of which values to include (True is include).
        """
        mask = np.full(data.shape, True)
        if self.min_value is not None:
            mask = data >= self.min_value

        if self.max_value is not None:
            mask *= data <= self.max_value

        if self._exclude_nodata and self.nodata_value is not None:
            mask = mask & (data != self.nodata_value)

        return mask

    def _value_mask(self, data, values, include=True):
        """
        Mask exclusion layer based on values to include or exclude

        Parameters
        ----------
        data : ndarray
            Exclusions data to create mask from
        values : list
            Values to include or exclude.
        include : boolean
            Flag as to whether values should be included or excluded.
            If True, output mask will be True where data == values.
            If False, output mask will be True where data != values.

        Returns
        -------
        mask : ndarray
            Boolean mask of which values to include (True is include)
        """
        mask = np.isin(data, values)

        if not include:
            mask = ~mask

        # only include if not nodata
        if self._exclude_nodata and self.nodata_value is not None:
            mask = mask & (data != self.nodata_value)

        return mask

    def _exclusion_mask(self, data):
        """
        Mask exclusion layer based on values to exclude

        Parameters
        ----------
        data : ndarray
            Exclusions data to create mask from

        Returns
        -------
        mask : ndarray
            Boolean mask of which values to include (True is include)
        """
        mask = self._value_mask(data, self.exclude_values, include=False)

        return mask

    def _inclusion_mask(self, data):
        """
        Mask exclusion layer based on values to include

        Parameters
        ----------
        data : ndarray
            Exclusions data to create mask from

        Returns
        -------
        mask : ndarray
            Boolean mask of which values to include (True is include)
        """
        mask = self._value_mask(data, self.include_values, include=True)

        return mask

    def _weights_mask(self, data):
        """
        Mask exclusion layer based on the weights for each inclusion value

        Parameters
        ----------
        data : ndarray
            Exclusions data to create mask from

        Returns
        -------
        mask : ndarray
            Percentage of value to include
        """
        mask = None
        for value, weight in self.include_weights.items():
            if isinstance(value, str):
                value = float(value)

            weight = np.array([weight], dtype='float32')

            if mask is None:
                mask = self._value_mask(data, [value], include=True) * weight
            else:
                mask += self._value_mask(data, [value], include=True) * weight

        return mask


class ExclusionMask:
    """
    Class to create final exclusion mask
    """

    FILTER_KERNELS = {
        'queen': np.array([[1, 1, 1],
                           [1, 1, 1],
                           [1, 1, 1]]),
        'rook': np.array([[0, 1, 0],
                          [1, 1, 1],
                          [0, 1, 0]])}

    def __init__(self, excl_h5, layers=None, min_area=None,
                 kernel='queen', hsds=False, check_layers=False):
        """
        Parameters
        ----------
        excl_h5 : str | list | tuple
            Path to one or more exclusions .h5 files
        layers : list | NoneType
            list of LayerMask instances for each exclusion layer to combine
        min_area : float | NoneType
            Minimum required contiguous area in sq-km
        kernel : str
            Contiguous filter method to use on final exclusion
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS
        check_layers : bool
            Run a pre-flight check on each layer to ensure they contain
            un-excluded values
        """
        self._layers = {}
        self._excl_h5 = ExclusionLayers(excl_h5, hsds=hsds)
        self._excl_layers = None
        self._check_layers = check_layers

        if layers is not None:
            if not isinstance(layers, list):
                layers = [layers]

            missing = [layer.name for layer in layers
                       if layer.name not in self.excl_layers]
            if any(missing):
                msg = ("ExclusionMask layers {} are missing from: {}"
                       .format(missing, self._excl_h5))
                logger.error(msg)
                raise KeyError(msg)

            for layer in layers:
                self.add_layer(layer)

        if kernel in ["queen", "rook"]:
            self._min_area = min_area
            self._kernel = kernel
            logger.debug('Initializing Exclusions mask with min area of {} '
                         'km2 and filter kernel "{}".'
                         .format(self._min_area, self._kernel))
        else:
            raise KeyError('kernel must be "queen" or "rook"')

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

        if type is not None:
            raise

    def __repr__(self):
        msg = ("{} from {} with {} input layers"
               .format(self.__class__.__name__, self.excl_h5.h5_file,
                       len(self)))

        return msg

    def __len__(self):
        return len(self.layers)

    def __getitem__(self, *ds_slice):
        """Get the multiplicative inclusion mask.

        Parameters
        ----------
        ds_slice : int | slice | list | ndarray
            What to extract from ds, each arg is for a sequential axis.
            For example, (slice(0, 64), slice(0, 64)) will extract a 64x64
            exclusions mask.

        Returns
        -------
        mask : ndarray
            Multiplicative inclusion mask with all layers multiplied together
            ("and" operation) such that 1 is included, 0 is excluded,
            0.5 is half.
        """
        return self._generate_mask(*ds_slice)

    def close(self):
        """
        Close h5 instance
        """
        self.excl_h5.close()

    @property
    def shape(self):
        """
        Get the exclusions shape.

        Returns
        -------
        shape : tuple
            (rows, cols) shape tuple
        """
        return self.excl_h5.shape

    @property
    def excl_h5(self):
        """
        Open ExclusionLayers instance

        Returns
        -------
        _excl_h5 : ExclusionLayers
        """
        return self._excl_h5

    @property
    def excl_layers(self):
        """
        List of available exclusion layers in exclusions .h5

        Returns
        -------
        _excl_layers : list
        """
        if self._excl_layers is None:
            self._excl_layers = self.excl_h5.layers

        return self._excl_layers

    @property
    def layer_names(self):
        """
        List of layers to combines

        Returns
        -------
        list
        """
        return self._layers.keys()

    @property
    def layers(self):
        """
        List of LayerMask instances for each exclusion layer to combine

        Returns
        -------
         list
        """
        return self._layers.values()

    @property
    def mask(self):
        """
        Inclusion mask for entire exclusion domain

        Returns
        -------
        ndarray
        """
        mask = self[...]
        return mask

    @property
    def latitude(self):
        """
        Latitude coordinates array

        Returns
        -------
        ndarray
        """
        return self.excl_h5[LATITUDE]

    @property
    def longitude(self):
        """
        Longitude coordinates array

        Returns
        -------
        ndarray
        """
        return self.excl_h5[LONGITUDE]

    def add_layer(self, layer, replace=False):
        """
        Add layer to be combined

        Parameters
        ----------
        layer : LayerMask
            LayerMask instance to add to set of layers to be combined
        """

        if layer.name not in self.excl_layers:
            msg = "{} does not exist in {}".format(layer.name, self._excl_h5)
            logger.error(msg)
            raise KeyError(msg)

        if layer.name in self.layer_names:
            msg = "{} is already in {}".format(layer.name, self)
            if replace:
                msg += " replacing existing layer"
                logger.warning(msg)
                warn(msg)
            else:
                logger.error(msg)
                raise ExclusionLayerError(msg)

        layer.nodata_value = self.excl_h5.get_nodata_value(layer.name)
        if self._check_layers:
            if not layer[self.excl_h5[layer.name]].any():
                msg = ("Layer {} is fully excluded!".format(layer.name))
                logger.error(msg)
                raise ExclusionLayerError(msg)

        self._layers[layer.name] = layer

    @property
    def nodata_lookup(self):
        """Get a dictionary lookup of the nodata values for each layer name.

        Returns
        -------
        nodata : dict
            Lookup keyed by layer name and values are nodata values for the
            respective layers.
        """
        nodata = {}
        for layer_name in self.layer_names:
            nodata[layer_name] = self.excl_h5.get_nodata_value(layer_name)

        return nodata

    @classmethod
    def _area_filter(cls, mask, min_area, excl_area, kernel='queen'):
        """
        Ensure the contiguous area of included pixels is greater than
        prescribed minimum in sq-km

        Parameters
        ----------
        mask : ndarray
            Inclusion mask
        min_area : float
            Minimum required contiguous area in sq-km
        kernel : str
            Kernel type, either 'queen' or 'rook'
        excl_area : float
            Area of each exclusion pixel in km^2, assumes 90m resolution

        Returns
        -------
        mask : ndarray
            Updated inclusion mask
        """
        s = cls.FILTER_KERNELS[kernel]
        labels, _ = ndimage.label(mask > 0, structure=s)
        l, c = np.unique(labels, return_counts=True)

        min_counts = np.ceil(min_area / excl_area)
        pos = c[1:] < min_counts
        bad_labels = l[1:][pos]

        mask[np.isin(labels, bad_labels)] = 0

        return mask

    def _increase_mask_slice(self, ds_slice, n=1):
        """Increase the mask slice, e.g. from 64x64 to 192x192, to help the
        contiguous area filter be more accurate.

        Parameters
        ----------
        ds_slice : tuple
            Two entry tuple with x and y slices. Anything else will be passed
            through unaffected.
        n : int
            Number of blocks to increase in each direction. For example,
            a 64x64 slice with n=1 will increase to 192x192
            (increases by 64xn in each direction).

        Returns
        -------
        new_slice : tuple
            Two entry tuple with x and y slices with increased dimensions.
        sub_slice : tuple
            Two entry tuple with x and y slices to retrieve the original
            slice out of the bigger slice.
        """
        new_slice = ds_slice
        sub_slice = (slice(None), slice(None))

        if isinstance(ds_slice, tuple) and len(ds_slice) == 2:
            y_slice = ds_slice[0]
            x_slice = ds_slice[1]
            if isinstance(x_slice, slice) and isinstance(y_slice, slice):
                y_diff = n * np.abs(y_slice.stop - y_slice.start)
                x_diff = n * np.abs(x_slice.stop - x_slice.start)

                y_new_start = int(np.max((0, (y_slice.start - y_diff))))
                x_new_start = int(np.max((0, (x_slice.start - x_diff))))

                y_new_stop = int(np.min((self.shape[0],
                                         (y_slice.stop + y_diff))))
                x_new_stop = int(np.min((self.shape[1],
                                         (x_slice.stop + x_diff))))

                new_slice = (slice(y_new_start, y_new_stop),
                             slice(x_new_start, x_new_stop))

                if y_new_start == y_slice.start:
                    y_sub_start = 0
                else:
                    y_sub_start = int(n * y_diff)
                if x_new_start == x_slice.start:
                    x_sub_start = 0
                else:
                    x_sub_start = int(n * x_diff)

                y_sub_stop = y_sub_start + y_diff
                x_sub_stop = x_sub_start + x_diff

                sub_slice = (slice(y_sub_start, y_sub_stop),
                             slice(x_sub_start, x_sub_stop))

        return new_slice, sub_slice

    def _generate_ones_mask(self, ds_slice):
        """
        Generate mask of all ones

        Parameters
        ----------
        ds_slice : tuple
            dataset slice of interest along axis 0 and 1

        Returns
        -------
        mask : ndarray
            Array of ones slices down by ds_slice
        """

        ones_shape = ()
        for i, s in enumerate(self.shape):
            if i < len(ds_slice):
                ax_slice = ds_slice[i]
                if np.issubdtype(type(ax_slice), np.integer):
                    ones_shape += (ax_slice,)
                else:
                    ax = np.arange(s, dtype=np.int32)
                    ones_shape += (len(ax[ax_slice]), )
            else:
                ones_shape += (s, )

        mask = np.ones(ones_shape, dtype='float32')

        return mask

    def _add_layer_to_mask(self, mask, layer, ds_slice, check_layers,
                           combine_func):
        """Add layer mask to full mask."""
        layer_mask = self._compute_layer_mask(layer, ds_slice, check_layers)
        if mask is None:
            return layer_mask

        return combine_func(mask, layer_mask, dtype='float32')

    def _compute_layer_mask(self, layer, ds_slice, check_layers=False):
        """Compute mask for single layer, including extent."""
        layer_mask = self._masked_layer_data(layer, ds_slice)
        layer_mask = self._apply_layer_mask_extent(layer, layer_mask, ds_slice)

        logger.debug('Computed exclusions {} for {}. Layer has average value '
                     'of {:.2f}.'
                     .format(layer, ds_slice, layer_mask.mean()))
        log_mem(logger, log_level='DEBUG')

        if check_layers and not layer_mask.any():
            msg = "Layer {} is fully excluded!".format(layer.name)
            logger.error(msg)
            raise ExclusionLayerError(msg)

        return layer_mask

    def _apply_layer_mask_extent(self, layer, layer_mask, ds_slice):
        """Apply extent to layer mask, if any."""
        if layer.extent is None:
            return layer_mask

        layer_extent = self._masked_layer_data(layer.extent, ds_slice)
        if not np.array_equal(layer_extent, layer_extent.astype(bool)):
            msg = ("Extent layer must be boolean (i.e. 0 and 1 values "
                   "only)! Please check your extent definition for layer "
                   "{} to ensure you are producing a boolean layer!"
                   .format(layer.name))
            logger.error(msg)
            raise ExclusionLayerError(msg)

        logger.debug("Filtering mask for layer %s down to specified extent",
                     layer.name)
        layer_mask = np.where(layer_extent, layer_mask, 1)
        return layer_mask

    def _masked_layer_data(self, layer, ds_slice):
        """Extract masked data for layer."""
        return layer[self.excl_h5[(layer.name, ) + ds_slice]]

    def _generate_mask(self, *ds_slice, check_layers=False):
        """
        Generate multiplicative inclusion mask from exclusion layers.

        Parameters
        ----------
        ds_slice : int | slice | list | ndarray
            What to extract from ds, each arg is for a sequential axis.
            For example, (slice(0, 64), slice(0, 64)) will extract a 64x64
            exclusions mask.
        check_layers : bool
            Check each layer as each layer is extracted to ensure they contain
            un-excluded values. This should only really be True if ds_slice is
            for the full inclusion mask. Otherwise, this could raise an error
            for a fully excluded mask for just one excluded SC point.

        Returns
        -------
        mask : ndarray
            Multiplicative inclusion mask with all layers multiplied together
            ("and" operation) such that 1 is included, 0 is excluded,
            0.5 is half.
        """

        mask = None
        ds_slice, sub_slice = self._parse_ds_slice(ds_slice)

        if self.layers:
            force_include = []
            for layer in self.layers:
                if layer.force_include:
                    force_include.append(layer)
                else:
                    mask = self._add_layer_to_mask(mask, layer, ds_slice,
                                                   check_layers,
                                                   combine_func=np.minimum)
            for layer in force_include:
                mask = self._add_layer_to_mask(mask, layer, ds_slice,
                                               check_layers,
                                               combine_func=np.maximum)

            if self._min_area is not None:
                mask = self._area_filter(mask, self._min_area,
                                         self._excl_h5.pixel_area,
                                         kernel=self._kernel)
                mask = mask[sub_slice]
        else:
            if self._min_area is not None:
                ds_slice = sub_slice

            mask = self._generate_ones_mask(ds_slice)

        return mask

    def _parse_ds_slice(self, ds_slice):
        """Parse a dataset slice to make it the proper dimensions and also
        optionally increase the dataset slice to make the contiguous area
        filter more accurate

        Parameters
        ----------
        ds_slice : int | slice | list | ndarray
            What to extract from ds, each arg is for a sequential axis.
            For example, (slice(0, 64), slice(0, 64)) will extract a 64x64
            exclusions mask.

        Returns
        -------
        ds_slice : tuple
            Two entry tuple with x and y slices with increased dimensions.
        sub_slice : tuple
            Two entry tuple with x and y slices to retrieve the original
            slice out of the bigger slice.
        """

        if len(ds_slice) == 1 & isinstance(ds_slice[0], tuple):
            ds_slice = ds_slice[0]

        sub_slice = None
        if self._min_area is not None:
            ds_slice, sub_slice = self._increase_mask_slice(ds_slice, n=1)

        return ds_slice, sub_slice

    @classmethod
    def run(cls, excl_h5, layers=None, min_area=None,
            kernel='queen', hsds=False):
        """
        Create inclusion mask from given layers

        Parameters
        ----------
        excl_h5 : str | list | tuple
            Path to one or more exclusions .h5 files
        layers : list | NoneType
            list of LayerMask instances for each exclusion layer to combine
        min_area : float | NoneType
            Minimum required contiguous area in sq-km
        kernel : str
            Contiguous filter method to use on final exclusion
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS

        Returns
        -------
        mask : ndarray
            Full inclusion mask
        """
        with cls(excl_h5, layers=layers, min_area=min_area,
                 kernel=kernel, hsds=hsds) as f:
            mask = f.mask

        return mask


class ExclusionMaskFromDict(ExclusionMask):
    """
    Class to initialize ExclusionMask from a dictionary defining layers
    """

    def __init__(self, excl_h5, layers_dict=None, min_area=None,
                 kernel='queen', hsds=False, check_layers=False):
        """
        Parameters
        ----------
        excl_h5 : str | list | tuple
            Path to one or more exclusions .h5 files
        layers_dict : dict | NoneType
            Dictionary of LayerMask arugments {layer: {kwarg: value}}
        min_area : float | NoneType
            Minimum required contiguous area in sq-km
        kernel : str
            Contiguous filter method to use on final exclusion
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS
        check_layers : bool
            Run a pre-flight check on each layer to ensure they contain
            un-excluded values
        """
        if layers_dict is not None:
            layers = []
            for layer, kwargs in layers_dict.items():
                layers.append(LayerMask(layer, **kwargs))
        else:
            layers = None

        super().__init__(excl_h5, layers=layers, min_area=min_area,
                         kernel=kernel, hsds=hsds, check_layers=check_layers)

    @classmethod
    def extract_inclusion_mask(cls, excl_fpath, tm_dset, excl_dict=None,
                               area_filter_kernel='queen', min_area=None):
        """
        Extract the full inclusion mask from excl_fpath using the given
        exclusion layers and whether or not to run a minimum area filter

        Parameters
        ----------
        excl_fpath : str | list | tuple
            Filepath to exclusions h5 with techmap dataset
            (can be one or more filepaths).
        tm_dset : str
            Dataset name in the techmap file containing the
            exclusions-to-resource mapping data.
        excl_dict : dict | None
            Dictionary of exclusion keyword arugments of the format
            {layer_dset_name: {kwarg: value}} where layer_dset_name is a
            dataset in the exclusion h5 file and kwarg is a keyword argument to
            the reV.supply_curve.exclusions.LayerMask class.
        area_filter_kernel : str, optional
            Contiguous area filter method to use on final exclusions mask,
            by default "queen"
        min_area : float, optional
            Minimum required contiguous area filter in sq-km,
            by default None

        Returns
        -------
        inclusion_mask : ndarray
            Pre-computed 2D inclusion mask (normalized with expected range:
            [0, 1], where 1 is included and 0 is excluded)
        """
        logger.info('Pre-extracting full exclusion mask, this could take '
                    'up to 30min for a large exclusion config...')
        with cls(excl_fpath, layers_dict=excl_dict, check_layers=False,
                 min_area=min_area, kernel=area_filter_kernel) as f:
            inclusion_mask = f._generate_mask(..., check_layers=True)
            tm_mask = f._excl_h5[tm_dset] == -1
            inclusion_mask[tm_mask] = 0

        logger.info('Finished extracting full exclusion mask.')
        logger.info('The full exclusion mask has {:.2f}% of area included.'
                    .format(100 * inclusion_mask.sum()
                            / inclusion_mask.size))

        if inclusion_mask.sum() == 0:
            msg = 'The exclusions inputs resulted in a fully excluded mask!'
            logger.error(msg)
            raise SupplyCurveInputError(msg)

        return inclusion_mask

    @classmethod
    def run(cls, excl_h5, layers_dict=None, min_area=None,
            kernel='queen', hsds=False):
        """
        Create inclusion mask from given layers dictionary

        Parameters
        ----------
        excl_h5 : str | list | tuple
            Path to one or more exclusions .h5 files
        layers_dict : dict | NoneType
            Dictionary of LayerMask arugments {layer: {kwarg: value}}
        min_area : float | NoneType
            Minimum required contiguous area in sq-km
        kernel : str
            Contiguous filter method to use on final exclusion
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS

        Returns
        -------
        mask : ndarray
            Full inclusion mask
        """
        with cls(excl_h5, layers_dict=layers_dict, min_area=min_area,
                 kernel=kernel, hsds=hsds) as f:
            mask = f.mask

        return mask


class FrictionMask(ExclusionMask):
    """Class to handle exclusion-style friction layer."""

    def __init__(self, fric_h5, fric_dset, hsds=False, check_layers=False):
        """
        Parameters
        ----------
        fric_h5 : str
            Path to friction layer .h5 file (same format as exclusions file)
        fric_dset : str
            Friction layer dataset in fric_h5
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS
        check_layers : bool
            Run a pre-flight check on each layer to ensure they contain
            un-excluded values
        """
        self._fric_dset = fric_dset
        L = [LayerMask(fric_dset, use_as_weights=True, exclude_nodata=False)]
        super().__init__(fric_h5, layers=L, min_area=None, hsds=hsds,
                         check_layers=check_layers)

    def _generate_mask(self, *ds_slice):
        """
        Generate multiplicative friction layer mask.

        Parameters
        ----------
        ds_slice : int | slice | list | ndarray
            What to extract from ds, each arg is for a sequential axis.
            For example, (slice(0, 64), slice(0, 64)) will extract a 64x64
            exclusions mask.

        Returns
        -------
        mask : ndarray
            Multiplicative friction layer mask with nodata values set to 1.
        """

        mask = None
        if len(ds_slice) == 1 & isinstance(ds_slice[0], tuple):
            ds_slice = ds_slice[0]

        layer_slice = (self._layers[self._fric_dset].name, ) + ds_slice
        mask = self._layers[self._fric_dset][self.excl_h5[layer_slice]]
        mask[(mask == self._layers[self._fric_dset].nodata_value)] = 1

        return mask

    @classmethod
    def run(cls, excl_h5, fric_dset, hsds=False):
        """
        Create inclusion mask from given layers dictionary

        Parameters
        ----------
        fric_h5 : str
            Path to friction layer .h5 file (same format as exclusions file)
        fric_dset : str
            Friction layer dataset in fric_h5
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS

        Returns
        -------
        mask : ndarray
            Full inclusion mask
        """
        L = [LayerMask(fric_dset, use_as_weights=True, exclude_nodata=False)]
        with cls(excl_h5, *L, min_area=None, hsds=hsds) as f:
            mask = f.mask

        return mask
